import inspect
import json
from ast import literal_eval
from datetime import datetime
from pathlib import Path
from time import sleep
from typing import Literal, Optional

from shadow_scholar.cli import Argument, cli, safe_import

from .constants import CSS, INSTRUCTIONS

with safe_import():
    import gradio as gr

    from .galai_model import Model


class ModelWrapper:
    def __init__(
        self,
        name: str,
        precision: Literal["full", "mixed", "byte"] = "full",
        tensor_parallel: bool = False,
        logdir: Optional[str] = None,
        leftover_space: float = 0.3,
        print_to_console: bool = True,
    ):
        self.model = Model(
            name=name,
            precision=precision,
            tensor_parallel=tensor_parallel,
            leftover_space=leftover_space,
        )
        self.start_time = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        self.logdir = Path(logdir) if logdir else None
        self.signature = inspect.signature(self.model.generate)
        self.print_to_console = print_to_console

    def __str__(self) -> str:
        return str(self.model) + f" start_time={self.start_time}"

    def log(self, arguments, output):
        out = {"input": arguments, "output": output}

        if self.print_to_console:
            print(json.dumps(out))

        if self.logdir is None:
            return

        self.logdir.mkdir(parents=True, exist_ok=True)

        fn = f'{self.model.name.replace("/", "_")}_{self.start_time}.jsonl'
        with open(self.logdir / fn, "a") as f:
            f.write(json.dumps(out) + "\n")

    def __call__(self, *args, **kwargs):
        arguments = self.signature.bind(*args, **kwargs).arguments

        if isinstance(opt := arguments.pop("extra_options", None), list):
            arguments["extra_options"] = {
                # evaluate strings as python literals
                k: literal_eval(v)
                for k, v in opt
                # no empty strings
                if k.strip() and v.strip()
            }

        arguments["top_k"] = (
            int(top_k) if (top_k := arguments.pop("top_k", None)) else None
        )

        arguments["top_p"] = (
            float(top_p) if (top_p := arguments.pop("top_p", None)) else None
        )

        arguments["penalty_alpha"] = (
            float(pa) if (pa := arguments.pop("penalty_alpha", None)) else None
        )

        output = self.model.generate(**arguments)
        self.log(arguments, output)
        return output


@cli(
    "app.galactica",
    arguments=[
        Argument(
            "-m",
            "--model-name",
            default="mini",
            choices=["mini", "base", "standard", "large", "huge"],
            help="Pretrained model or path to local checkpoint",
        ),
        Argument(
            "-p",
            "--server-port",
            default=7860,
            help="Port to run the server on",
        ),
        Argument(
            "-n",
            "--server-name",
            default="localhost",
            help="Server address to run the gradio app at",
        ),
        Argument(
            "-e",
            "--precision",
            default="full",
            help="Precision to use for the model",
            choices=["full", "mixed", "byte"],
        ),
        Argument(
            "-a",
            "--parallelize",
            default=False,
            action="store_true",
            help="Parallelize the model across multiple GPUs.",
        ),
        Argument(
            "-l",
            "--logdir",
            default=None,
            help="Directory to log inputs and outputs to",
        ),
        Argument(
            "-s",
            "--leftover-space",
            default=0.3,
            type=float,
            help="the amount of extra space to leave on the gpu during",
        ),
    ],
    requirements=[
        "gradio",
        "transformers",
        "torch",
        "accelerate",
        "psutil",
    ],
)
def run_galactica_demo(
    model_name: str = "mini",
    server_port: int = 7860,
    server_name: str = "localhost",
    precision: Literal["full", "mixed", "byte"] = "full",
    parallelize: bool = False,
    logdir: Optional[str] = None,
    leftover_space: float = 0.3,
):
    gl_model = ModelWrapper(
        name=model_name,
        precision=precision,
        tensor_parallel=parallelize,
        logdir=logdir,
        leftover_space=leftover_space,
    )

    with gr.Blocks(css=CSS) as demo:
        with gr.Row():
            gr.Markdown(f"# Galactica {model_name.capitalize()} Demo")
        with gr.Tab("Demo"):
            with gr.Row():
                gr.Markdown(
                    f"**Currently loaded**: {gl_model}\n\n"
                    + "Available models:\n"
                    + "\n".join(
                        f" - {k.capitalize()}: `{n}`"
                        for k, (n, _) in gl_model.model.variants.items()
                    )
                )
            with gr.Row():
                with gr.Column():
                    input_text = gr.Textbox(
                        lines=5, label="Input", placeholder="Prompt text"
                    )
                    submit_button = gr.Button(label="Generate")

                    max_new_tokens = gr.Number(
                        value=128,
                        label=(
                            "max_new_tokens: max number of new tokens "
                            "the model should generate"
                        ),
                    )
                    top_p = gr.Textbox(
                        value="",
                        label=(
                            "top_p: if set to float < 1, only the "
                            "smallest set of most probable tokens with "
                            "probabilities that add up to top_p or higher "
                            "are kept for generation."
                        ),
                    )
                    top_k = gr.Textbox(
                        value="",
                        label=(
                            "top_k: size of the candidate set that is "
                            "used to re-rank for contrastive search"
                        ),
                    )
                    penalty_alpha = gr.Textbox(
                        value="",
                        label=(
                            "penalty_alpha: degeneration penalty for "
                            "contrastive search"
                        ),
                    )

                    num_beams = gr.Number(
                        value=1,
                        label=(
                            "num_beams: number of beams for beam search. "
                            "1 means no beam search."
                        ),
                    )
                    num_return_sequences = gr.Number(
                        value=1,
                        label=(
                            "num_return_sequences: number of separate"
                            "computed returned sequences for each element "
                            "in the batch."
                        ),
                    )
                    return_full_text = gr.Checkbox(
                        label=(
                            "return_full_text: whether to return the full "
                            "text or just the newly generated text."
                        ),
                        value=True,
                    )
                    new_doc = gr.Checkbox(
                        label=(
                            "new_doc: whether the model should attempt "
                            "to generate a full document"
                        ),
                        value=False,
                    )
                    extra_options = gr.Dataframe(
                        label="Extra options to pass to model.generate()",
                        headers=["Parameter", "Value"],
                        col_count=2,
                        type="array",
                        interactive=True,
                    )

                with gr.Column():
                    output_text = gr.Textbox(
                        lines=25,
                        label="Output",
                        placeholder="Generated text",
                        interactive=False,
                    )

            submit_button.click(
                fn=gl_model,
                inputs=[
                    input_text,
                    max_new_tokens,
                    new_doc,
                    top_p,
                    top_k,
                    penalty_alpha,
                    num_beams,
                    num_return_sequences,
                    return_full_text,
                    extra_options,
                ],
                outputs=[output_text],
            )
        with gr.Tab("Instructions"):
            with gr.Row():
                gr.Markdown(INSTRUCTIONS, elem_id="instructions")

    try:
        demo.queue(concurrency_count=1)
        demo.launch(server_name=server_name, server_port=server_port)
    except Exception:
        sleep(1)
        gr.close_all()



import json
import os
from typing import List, Literal, Optional

from shadow_scholar.cli import Argument, cli, load_kwargs

from .datasets import Document, Qrel, Query, dataset_registry
from .rankers import ranker_registry
from .slicers import slicer_registry


@cli(
    name="app.pdod",
    arguments=[
        Argument(
            "-rn",
            "--ranker-name",
            help=f"Ranker to use. Options: {ranker_registry.keys()}",
            required=True,
        ),
        Argument(
            "-rk",
            "--ranker-kwargs",
            help="JSON-encoded kwargs for the ranker",
            type=load_kwargs,
        ),
        Argument(
            "-dn",
            "--dataset-name",
            help=f"Dataset to use. Options: {dataset_registry.keys()}",
        ),
        Argument(
            "-dp",
            "--docs-path",
            help=(
                "Path to file or directory containing a JSONL of documents. "
                f"Documents are expect to have fields: {Document.fields()}"
            ),
        ),
        Argument(
            "-qp",
            "--queries-path",
            help=(
                "Path to file or directory containing a JSONL of queries. "
                f"Queries are expect to have fields: {Query.fields()}"
            ),
        ),
        Argument(
            "-rp",
            "--qrels-path",
            help=(
                "Path to file or directory containing a JSONL of qrels. "
                f"Qrels are expect to have fields: {Qrel.fields()}"
            ),
        ),
        Argument(
            "-qm",
            "--qrels-mode",
            help="Whether to use open or judged qrels when qrels are provided",
            choices=["open", "judged"],
            default="judged",
        ),
        Argument(
            "-op",
            "--output-path",
            help="If provided, path to write results to",
        ),
        Argument(
            "-sl",
            "--slicer-name",
            help=(
                f"Slicer to use. Options: {slicer_registry.keys()} "
                "If not provided, no slicing will be performed."
            ),
        ),
        Argument(
            "-sk",
            "--slicer-kwargs",
            help="JSON-encoded kwargs for the slicer",
            type=load_kwargs,
        ),
        Argument(
            "-ms",
            "--metrics",
            help="Metrics to use for evaluation",
            type=list,
            default=None,
        ),
    ],
    requirements=[
        "numpy",
        "spacy",
        "torch",
        ("scikit-learn", "sklearn"),
        "transformers",
        "unidecode",
        ("sentence-transformers", "sentence_transformers"),
    ],
)
def run_pdod(
    ranker_name: str,
    ranker_kwargs: Optional[dict] = None,
    dataset_name: Optional[str] = None,
    docs_path: Optional[str] = None,
    queries_path: Optional[str] = None,
    qrels_path: Optional[str] = None,
    qrels_mode: Literal["open", "judged"] = "judged",
    output_path: Optional[str] = None,
    slicer_name: Optional[str] = None,
    slicer_kwargs: Optional[dict] = None,
    metrics: Optional[List[str]] = None,
):
    metrics = metrics or []

    if docs_path:
        dataset = dataset_registry.get("from_files")(
            docs_path=docs_path,
            queries_path=queries_path,
            qrels_path=qrels_path,
        )
    elif dataset_name:
        dataset = dataset_registry.get(dataset_name)()
    else:
        raise ValueError("Either dataset_name or docs_path must be provided")

    is_interactive = len(dataset.queries) == 0 and queries_path is None
    is_evaluation = len(dataset.qrels) > 0

    if slicer_name:
        slicer = slicer_registry.get(slicer_name)(**(slicer_kwargs or {}))
        dataset.documents = slicer.slice(dataset.documents)

    ranker = ranker_registry.get(ranker_name)(**(ranker_kwargs or {}))
    ranker.warmup(dataset.documents)
    output: List[Document] = []

    while True:
        if is_interactive:
            # interactive mode, ask user for query
            try:
                query_text = input("Query: ")
            except KeyboardInterrupt:
                print("Exiting...")
                break
            dataset.queries.append(Query(text=query_text, qid="_"))
        if len(dataset.queries) == 0:
            # no more queries, we're done
            break

        # take the next query
        query = dataset.queries.pop(0)

        if is_evaluation and qrels_mode == "judged":
            docs = dataset.iter_close(query)
        elif is_evaluation and qrels_mode == "open":
            docs = dataset.iter_open(query)
        else:
            docs = dataset.documents

        scored_docs = ranker.score(query.text, docs)

        if is_interactive:
            width = os.get_terminal_size().columns
            # only show the top 5 results
            for doc in scored_docs[:5]:
                print(f"{doc.score:.3f} {doc.text[:width - 10]}...")
            print()
        else:
            scored_docs = [doc.query(query.qid) for doc in scored_docs]

        output.extend(scored_docs)

    if output_path is not None:
        with open(output_path, "w") as f:
            for doc in output:
                f.write(json.dumps(doc.as_dict()) + "\n")

    if len(dataset.qrels) > 0:
        raise NotImplementedError("Evaluation not yet implemented")


@cli(name="app.pdod.web")
def run_pdod_web_ui():
    raise NotImplementedError("Web UI not yet implemented")



import os
from dataclasses import dataclass, field
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Literal,
    NamedTuple,
    Optional,
    Tuple,
    Union,
)

import openai
from jinja2 import Template

from .library import Paper, clean_html
from .logging import get_logger

OPENAI_API_KEY = os.environ.get("OPENAI_API_KEY", None)


class ChatMessage(NamedTuple):
    role: Literal["user", "system", "assistant"]
    content: Template

    def to_json(self, content: Optional[dict] = None) -> Dict[str, str]:
        return {
            "role": self.role,
            "content": self.content.render(**(content or {})),
        }


class OpenAiPrompt:
    def __init__(
        self,
        prompt: Union[str, List[Tuple[str, str]]],
        openai_api_key: Optional[str] = OPENAI_API_KEY,
        model: str = "text-davinci-003",
        mode: Optional[Literal["chat", "completion"]] = None,
        temperature: float = 0.5,
        max_tokens: int = 256,
    ):
        if openai_api_key is None:
            raise ValueError("No OpenAI API key provided")
        openai.api_key = openai_api_key

        self.mode = mode or (
            "chat" if model in {"gpt-4", "gpt-3.5-turbo"} else "completion"
        )
        self.model = model

        prompt = [("user", prompt)] if isinstance(prompt, str) else prompt
        self.prompt = [
            ChatMessage(role=role, content=Template(content))  # type: ignore
            for role, content in prompt
        ]

        self.temperature = temperature
        self.max_tokens = max_tokens

        self.logger = get_logger(__name__, disable_cloudwatch="yes")

    def to_json(self) -> Dict[str, Any]:
        return {
            "model": self.model,
            "mode": self.mode,
            "prompt": [message.to_json() for message in self.prompt],
            "temperature": self.temperature,
            "max_tokens": self.max_tokens,
        }

    def completion(self, content: dict, params: Optional[dict] = None) -> str:
        request = {
            "model": self.model,
            "prompt": self.prompt[0].content.render(**content),
            "temperature": self.temperature,
            "max_tokens": self.max_tokens,
            **(params or {}),
        }
        response = openai.Completion.create(**request)
        self.logger.info(
            {
                "request": request,
                "response": response,
                "endpoint": "Completion",
                "action": "completion_openai_llm",
            }
        )
        return response.choices[0].text  # pyright: ignore

    def chat(self, content: dict, params: Optional[dict] = None) -> str:
        request = {
            "model": self.model,
            "messages": [message.to_json(content) for message in self.prompt],
            "temperature": self.temperature,
            "max_tokens": self.max_tokens,
            **(params or {}),
        }
        response = openai.ChatCompletion.create(**request)
        self.logger.info(
            {
                "request": request,
                "response": response,
                "endpoint": "ChatCompletion",
                "action": "chat_openai_llm",
            }
        )
        return response.choices[0].message.content  # pyright: ignore

    def __call__(self, **content: Any) -> str:
        if self.mode == "chat":
            return self.chat(content)
        elif self.mode == "completion":
            return self.completion(content)
        else:
            raise ValueError(f"Unknown mode: {self.mode}")


class Summary(NamedTuple):
    summary: str
    score: float
    paper: "Paper"

    @property
    def relevant(self) -> bool:
        return self.score > 0.5

    @property
    def id(self) -> str:
        return self.paper.id

    def html_format_id(self) -> str:
        return self.paper.html_format_id()

    def html_format_summary(self) -> str:
        body = "".join([f"<p>{line}</p>" for line in self.summary.split("\n")])
        title = self.paper.html_format_title()
        return f"{title}{body}"

    def format_relevant(self) -> str:
        return "✅" if self.relevant else "❌"

    def to_json(self) -> Dict[str, Any]:
        return {
            "summary": self.summary,
            "score": self.score,
            "paper": self.paper.to_json(),
        }

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> "Summary":
        return cls(
            summary=data["summary"],
            score=data["score"],
            paper=Paper.from_json(data["paper"]),
        )


@dataclass
class Summaries:
    summaries: Dict[Tuple[str, str], Summary] = field(default_factory=dict)

    def __contains__(self, id_: str) -> bool:
        return id_ in self.summaries

    def iter(self, prompt: Optional[str] = None) -> Iterable[Summary]:
        for (_, summary_prompt), summary in self.summaries.items():
            if prompt is None or summary_prompt == prompt:
                yield summary

    def append(
        self,
        summary: Union[str, Summary],
        prompt: str = "",
        score: Optional[float] = None,
        paper: Optional[Paper] = None,
    ):
        if isinstance(summary, str):
            if score is None or paper is None:
                raise ValueError("Must provide score/paper if summary is str")
            summary = Summary(
                summary=clean_html(summary), score=score, paper=paper
            )

        self.summaries[(summary.id, prompt)] = summary

    def filter(self):
        for id_ in list(self.summaries):
            if not self.summaries[id_].relevant:
                del self.summaries[id_]

    def table(self, prompt: Optional[str] = None) -> List[List[str]]:
        if len(self.summaries) == 0:
            # must return empty list to avoid error
            return [["", "", ""]]

        return [
            [s.html_format_id(), s.html_format_summary(), s.format_relevant()]
            for s in sorted(
                list(self.iter(prompt)),
                key=lambda s: s.paper.short_id or -s.paper.score,
            )
        ]

    def to_json(self) -> Dict[str, Any]:
        return {
            "summaries": {
                f"{id_}|||{prompt}": summary.to_json()
                for (id_, prompt), summary in self.summaries.items()
            }
        }

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> "Summaries":
        summaries = {}
        for key, value in data["summaries"].items():
            id_, prompt = key.split("|||", 1)
            summaries[(id_, prompt)] = Summary.from_json(value)
        return cls(summaries=summaries)



# flake8: noqa

CSS = """
code {
    white-space : pre-wrap !important;
}
"""


INSTRUCTIONS = '''\
The following instructions are from the README on the [Galactica GitHub repo](https://github.com/paperswithcode/galai).

## Capabilities

GALACTICA is a stand-alone LM which is not instruction tuned. Because of this you need to use the correct prompts to get good results. In this note, we go over some of the special tokens, and prompt styles you will need to use to get good results.

We demonstrate some examples using the standard (6.7B) model below.

📚 **Predict Citations**:

You need to use `[START_REF]`:

```python
model.generate("The Transformer architecture [START_REF]")
# The Transformer architecture [START_REF] Attention is All you Need, Vaswani[END_REF] is a sequence-to-sequence model that uses self-attention to capture long-range dependencies between input and output tokens. The Transformer has been shown to achieve state-of-the-art results on a wide range of natural
```

🔢 **Predict LaTeX**:

```python
model.generate("The Schwarzschild radius is defined as: \\[")
# The Schwarzschild radius is defined as: \\[r_{s}=\\frac{2GM}{c^{2}}\\]\\n\\nwhere \\(G\\) is the gravitational constant, \\(M\\) is the mass of the black hole, and
```

🤔 **Reasoning**:

Reasoning uses the special `<work>` token:

```python
model.generate("A force of 0.6N is applied to an object, which accelerates at 3m/s. What is its mass? <work>")
# What force should be applied to accelerate an object of mass 3kg to 10m/s? <work>\\nWe can use Newton's second law: F = ma. We can substitute variables to get:\\n\\n\\[ F = \\left(66kg
```

⚛️ **Generate Molecules**:

```python
model.generate("[START_I_SMILES]", max_length=200)
# [START_I_SMILES]CCC1=CC=C(C=C1)C(=O)NC2=CC=CC(=C2)C(=O)NC3=CC=C(C=C3)S(=O)(=O)N[END_I_SMILES]\\n\\n### Molecular Formula\\n\\nC22H21N3O4S\\n\\n## Chemical and Physical Properties\\n\\nThe following are chemical properties for 3-[[3-(4-ethylphenyl)-3-oxo-propanoyl]amino]-N-(4-sulfamoylphenyl)benzamide.\\n\\n### Computed Properties\\n\\n| Property Name | Property Value\\n| --- | ----------- |\\n| Molecular Weight | 423.5\\n| XLogP3-AA Log P | 3.2\\n| Hydrogen Bond Donor Count | 3\\n| Hydrogen Bond Acceptor Count
```

🧑‍🔬 **Predict Protein Annotations**:

```python
model.generate("[START_AMINO]GHMQSITAGQKVISKHKNGRFYQCEVVRLTTETFYEVNFDDGSFSDNLYPEDIVSQDCLQFGPPAEGEVVQVRWTDGQVYGAKFVASHPIQMYQVEFEDGSQLVVKRDDVYTLDEELP[END_AMINO] ## Keywords", max_length=200)
# '[START_AMINO]GHMQSITAGQKVISKHKNGRFYQCEVVRLTTETFYEVNFDDGSFSDNLYPEDIVSQDCLQFGPPAEGEVVQVRWTDGQVYGAKFVASHPIQMYQVEFEDGSQLVVKRDDVYTLDEELP[END_AMINO] ## Keywords\\n\\nCytoplasm, Methyltransferase, rRNA processing, S-adenosyl-L-methionine, Transferase\\n\\n## References\\n\\nQuestion: What are some articles for Ribosomal RNA small subunit methyltransferase H?\\n\\nAnswer: \\n\\n[START_REF] Comparative Genomics of 28 Salmonella enterica Isolates: Evidence for CRISPR-Mediated Adaptive Sublineage Evolution, Fricke[END_REF]\\n\\n</s>'
```

🖱️ **Free-Form Generation**

If you want autocomplete based functionality, it is often good to experiment with turning off `new_doc=True`. This makes it more likely for the model to think it is in the middle of a document, as opposed to the beginning.

```python
model.generate("The reason why Transformers replaced RNNs was because", new_doc=False)
# The reason why Transformers replaced RNNs was because they were able to capture long-term dependencies in the input sequence.\\n\\n# 2.2.2. Attention Mechanism\\n\\nThe attention mechanism was introduced in [START_REF] Neural Machine Translation by Jointly Learning to Align and Translate, Bahdan
```

❓ **Question Answering**

In the paper we prefix questions with "Q:" or "Question:". A typical format is "Question: question.\\n\\nAnswer:", for example:

```python
model.generate("Question: What is the notch signaling pathway?\\n\\nAnswer:")
# 'Question: What is the notch signaling pathway?\\n\\nAnswer: \\n\\nNotch signaling pathway is a cell-cell communication pathway that regulates cell fate decisions during development. It is involved in cell proliferation, differentiation, apoptosis, and cell migration. The Notch signaling pathway is activated by the binding of'
```

📄 **Documents**

When starting a document, you must use the start document token for good results. To do this, set `new_doc=True` in generate:

For some article types, like Wikipedia style articles, lecture notes and GitHub repositories, use `#` to begin, e.g:

```python
model.generate("# Multi-Head Attention\\n\\n", new_doc=True)
# # Multi-Head Attention\\n\\nThe multi-head attention mechanism is a generalization of the single-head attention mechanism. The multi-head attention mechanism is a combination of multiple single-head attention mechanisms. The multi-head attention mechanism is shown in Figure 2.\\n\\nThe multi-
```

For paper documents, use Title, e.g:

```python
model.generate("Title: Self-Supervised Learning, A Survey\\n\\nAuthors: John Smith\\n\\n", new_doc=True)
# Title: Self-Supervised Learning, A Survey\\n\\nAuthors: John Smith\\n\\n# Abstract\\n\\nSelf-supervised learning is a class of machine learning methods that learn representations of data without the need for human-provided labels.\\nIn this survey, we provide a comprehensive overview of the field
```

You can also try alternative sampling techniques for less repetitions, e.g.

```python
model.generate("Lecture 1: The Ising Model\\n\\n", new_doc=True, top_p=0.7, max_length=200)
# 'Lecture 1: The Ising Model\\n\\n# 13 Introduction\\n\\nWe will now look at a simple model for magnetism, the Ising model, which is\\na lattice model in which we consider only two spin values, up or down, and\\nwe want to understand how these spins interact with each other and how\\nthey get arranged in a particular state.\\n\\nWe will first consider the one-dimensional case, and then move on to\\nthe case of two-dimensional lattices, and then to higher dimensions.\\n\\n# 14 The One-Dimensional Ising Model\\n\\n# 14.1 The Model\\n\\nThe one-dimensional Ising model is the simplest case of the model, in\\nwhich the lattice is a line of \\(N\\) spins, each with two possible spin\\nvalues, up or down. In other words, we consider a line of \\(N\\) spins\\nwhere each spin can point up or down'
```

📜 **Summarization**

You can add "TLDR:" for TLDR summaries:

```python
TEXT = """Information overload is a major obstacle to scientific progress. The explosive growth in scientific literature and data has made it ever harder to discover useful insights in a large mass of information. Today scientific knowledge is accessed through search engines, but they are unable to organize scientific knowledge alone. In this paper we introduce Galactica: a large language model that can store, combine and reason about scientific knowledge. We train on a large scientific corpus of papers, reference material, knowledge bases and many other sources. We outperform existing models on a range of scientific tasks. On technical knowledge probes such as LaTeX equations, Galactica outperforms the latest GPT-3 by 68.2% versus 49.0%. Galactica also performs well on reasoning, outperforming Chinchilla on mathematical MMLU by 41.3% to 35.7%, and PaLM 540B on MATH with a score of 20.4% versus 8.8%. It also sets a new state-of-the-art on downstream tasks such as PubMedQA and MedMCQA dev of 77.6% and 52.9%. And despite not being trained on a general corpus, Galactica outperforms BLOOM and OPT-175B on BIG-bench. We believe these results demonstrate the potential for language models as a new interface for science. We open source the model for the benefit of the scientific community."""

model.generate(TEXT + "\\n\\nTLDR:", max_length=400)
# ...TLDR: We introduce Galactica, a large language model that can store, combine and reason about scientific knowledge.</s>
```

💎 **Entity extraction**

You can extract entities from documents. We use the abstract example (`TEXT`) from the previous section, and add questions

```python
ENT_TEXT = TEXT + '\\n\\nWhat scientific entities are mentioned in the abstract above?\\n\\n'

model.generate(ENT_TEXT, max_length=400)
# ...What scientific entities are mentioned in the abstract above?\\n\\nA: LaTeX equations, mathematical MMLU, MATH, PubMedQA, MedMCQA, BIG-bench</s>
```

👨‍🔬 **IUPAC Name prediction**

For this task, we used a prompt based off the PubChem document and prompted for the completion. We use the 6.7bn model for below:

```python
context = "[START_I_SMILES]C(C(=O)O)N[END_I_SMILES]\\n\\n## Chemical and Physical Properties\\n\\nThe following are chemical properties for"
model.generate(context, max_length=400)
# [START_I_SMILES]C(C(=O)O)N[END_I_SMILES]\\n\\n## Chemical and Physical Properties\\n\\nThe following are chemical properties for 2-amino-2-oxo-acetic acid
# Note this is an incorrect prediction
```

## Citation

```bibtex
@inproceedings{GALACTICA,
    title={GALACTICA: A Large Language Model for Science},
    author={Ross Taylor and Marcin Kardas and Guillem Cucurull and Thomas Scialom and Anthony Hartshorn and Elvis Saravia and Andrew Poulton and Viktor Kerkez and Robert Stojnic},
    year={2022}
}\
'''



from shadow_scholar.cli import safe_import

with safe_import():
    from parallelformers.policies.base import Layer, Policy
    from parallelformers.utils.dist_utils import AllReduceLinear
    from transformers.models.opt.modeling_opt import OPTDecoderLayer


__all__ = ["OPTDecoderLayerPolicyNoBias"]


class OPTDecoderLayerPolicyNoBias(Policy):
    @staticmethod
    def replace_arguments(config, world_size):
        return {
            "self_attn.embed_dim": config.hidden_size // world_size,
            "self_attn.num_heads": config.num_attention_heads // world_size,
        }

    @staticmethod
    def attn_qkv():
        return [
            Layer(
                weight="self_attn.q_proj.weight",
            ),
            Layer(
                weight="self_attn.k_proj.weight",
            ),
            Layer(
                weight="self_attn.v_proj.weight",
            ),
        ]

    @staticmethod
    def attn_out():
        return [
            Layer(
                weight="self_attn.out_proj.weight",
                replace=AllReduceLinear,
            ),
        ]

    @staticmethod
    def mlp_in():
        return [
            Layer(
                weight="fc1.weight",
            ),
        ]

    @staticmethod
    def mlp_out():
        return [
            Layer(
                weight="fc2.weight",
                replace=AllReduceLinear,
            ),
        ]

    @staticmethod
    def original_layer_class():
        return OPTDecoderLayer



import os
import re
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import List, Tuple, Union

from shadow_scholar.cli import Argument, cli, safe_import

with safe_import():
    import gradio as gr

    from .library import Paper, Stack
    from .llm import OpenAiPrompt, Summaries
    from .logging import get_logger
    from .search import BaseSearchEndpoint, GoogleSearchEndpoint, Queries


ConversationType = List[Tuple[Union[str, None], Union[str, None]]]


SEARCH_URI = "https://api.semanticscholar.org/graph/v1/paper/search/"
CSS_URI = Path(__file__).parent / "res" / "style.css"
LOGO_URI = Path(__file__).parent / "res" / "logo.svg"
JAVASCRIPT_URI = Path(__file__).parent / "res" / "script.js"

DESCRIPTION = """\
<div id="description">
    <div id="description-header">
        <h1>
            <span id="header-logo">{svg_logo}</span>
            <span id="header-text">Talk to Shadow Scholar</span>
        </h1>
    </div>
    <p>
        This is a demo of a chatbot that uses the <a target="_blank" \
        href="https://semanticscholar.org/">Semantic Scholar</a> API to \
        answer questions about scientific papers.You can either provide \
        it a list of papers to use, or ask the virtual scholar to search \
        papers for you.
    </p>
    <p>
        This virtual scholar is powered by the <a target="_blank" \
        href="https://platform.openai.com/docs/models/gpt-3">OpenAI API \
        </a> and uses the <code>{model}</code> model. It is intended as a \
        research prototype, <b>do not use it to power any user-facing \
        applications</b>.
    </p>
</div>
"""


QUERY_EXTRACTION_TEMPLATE = """\
I need help finding academic papers about the following topic: "{{ prompt }}".

I have access to an intelligent search engine that can find papers on the \
topic above. Could you suggest one or more queries to use? A few rules:
- Write them as a list, starting with "-".
- Suggest at most {% if cnt > 1 %}{{ cnt }} queries\
{% else %}one query{% endif %}.
- Avoid repetitive queries.
- Shorter queries are better.
- Do not use boolean operators, such as AND, OR, NOT, in the queries.
"""


ANSWER_FORMULATION_TEMPLATE = """\
I am interested in learning more about the following question: "{{ prompt }}".

Using the following evidence summaries I have summarized from papers:
{% for summary in summaries %}\
[{{ summary.paper.short_id }}] {{ summary.summary }}
{% endfor %}\

Could you answer the question? A few rules:
- Please cite the relevant snippet(s) using the format [n] \
(for example, [{{ summaries[0].paper.short_id }}] for the first snippet).
- Do not write a claim unless it appears in one of the snippets.
- If a snippet is not relevant, do not use it.
- You must be as brief and concise as possible; do not write more than \
five sentences.
- You must explain your reasoning.
- If none of the snippets contain a good answer, say that you can't answer \
the question.
"""

RESULTS_SUMMARIZATION_TEMPLATE = """\
I am interested in learning more about the following topic: \
"{{ prompt }}".

Does the following text contain any relevant information on the topic?

{{ paper.title }} ({{ paper.year }})
{{ " ".join(paper.authors) }}
{{ paper.abstract }}

If it does contain relevant information, write down one short sentences \
explaining why; do not start with "relevant". If the passage is not relevant, \
respond with "not relevant".
"""


def search_and_table(
    state: "State",
    max_results: Union[int, float],
    endpoint: "BaseSearchEndpoint",
) -> Tuple[List[List[str]], "State"]:
    searches = []
    max_results = int(max_results)
    for query in state.queries.use(max_results):
        results = endpoint(query, max_results=max_results)
        state.stack.extend(results)
        searches.append({"query": query, "results": [p.id for p in results]})

    state.stack.fetch()

    state.logger.info(
        {
            "searches": searches,
            "action": "search_and_table",
            "endpoint": endpoint.to_json(),
        }
    )

    return state.stack.table(), state


def add_user_paper_to_chatbot(
    paper_id: str, state: "State"
) -> Tuple[str, List[List[str]], "State"]:
    paper = (
        Paper.from_url(url=paper_id)
        if paper_id.startswith("http")
        else Paper(id=paper_id)
    )
    if paper is not None:
        state.stack.append(paper)
        state.stack.fetch()

    state.logger.info(
        {
            "action": "add_user_paper_to_chatbot",
            "paper_id": paper_id,
            "paper": paper.id if paper is not None else None,
        }
    )

    return "", state.stack.table(), state


def add_user_query_to_chatbot(
    query: str, state: "State"
) -> Tuple[str, ConversationType, "State"]:
    if query := query.strip():
        state.queries.add([(query, None)])

    state.logger.info(
        {
            "action": "add_user_query_to_chatbot",
            "query": query,
        }
    )
    return "", state.queries.list(), state


def make_queries_from_prompt(
    user_prompt: str,
    cnt_query: Union[int, float],
    state: "State",
    model: "OpenAiPrompt",
) -> Tuple[ConversationType, "State"]:
    if not (user_prompt := user_prompt.strip()):
        return state.queries.list(), state

    completion = model(prompt=user_prompt, cnt=int(cnt_query))
    extracted_queries = [
        strip_and_remove_quotes(q) for q in completion.split("\n") if q.strip()
    ]
    state.queries.add([(None, q) for q in extracted_queries])

    state.logger.info(
        {
            "action": "make_queries_from_prompt",
            "user_prompt": user_prompt,
            "cnt_query": cnt_query,
            "queries": extracted_queries,
        }
    )
    return state.queries.list(), state


def strip_and_remove_quotes(text: str) -> str:
    text = re.sub(r"^\-\s*", "", text)
    text = re.sub(r"[\"\']+", "", text)
    text = re.sub(r"(^\s+|\s+$)", "", text)
    return text


def summarize_results(
    prompt: str, state: "State", model: "OpenAiPrompt"
) -> Tuple[List[List[str]], "State"]:
    if not (prompt := prompt.strip()):
        return state.summaries.table(), state

    for paper in state.stack.papers.values():
        if paper.id in state.summaries:
            continue
        summary = model(paper=paper, prompt=prompt)
        not_relevant = re.search(
            r"^((non|not) relevant|irrelevant)\.*?\s?", summary, re.I
        )
        score = 0.0 if not_relevant else 1.0
        summary = re.sub(
            r"(^relevant.*?)([a-zA-Z])", r"\2", summary, flags=re.I
        )
        state.summaries.append(
            summary=summary, prompt=prompt, score=score, paper=paper
        )

    state.logger.info(
        {
            "action": "summarize_results",
            "prompt": prompt,
            "summaries": [
                {"paper": s.id, "summary": s.summary}
                for s in state.summaries.iter(prompt)
            ],
        }
    )
    return state.summaries.table(), state


def compute_answer_from_summaries(
    prompt: str,
    state: "State",
    model: "OpenAiPrompt",
) -> Tuple[str, "State"]:
    if not (prompt := prompt.strip()):
        return "Please provide a prompt", state

    relevant_summaries = [
        s for s in state.summaries.iter(prompt) if s.relevant
    ]

    if not relevant_summaries:
        return "No relevant summaries found", state

    answer = model(prompt=prompt, summaries=relevant_summaries)

    state.logger.info(
        {
            "action": "compute_answer_from_summaries",
            "prompt": prompt,
            "answer": answer,
        }
    )

    return answer, state


def run_all_fn(
    prompt: str,
    cnt_query: int,
    max_results: int,
    state: "State",
    search_endpoint: "BaseSearchEndpoint",
    query_prompt_model: "OpenAiPrompt",
    answer_prompt_model: "OpenAiPrompt",
    results_summarization: "OpenAiPrompt",
):
    state = State.new()

    # make queries from prompt
    queries, state = make_queries_from_prompt(
        user_prompt=prompt,
        cnt_query=cnt_query,
        state=state,
        model=query_prompt_model,
    )

    # search for papers
    results, state = search_and_table(
        state=state,
        max_results=max_results,
        endpoint=search_endpoint,
    )

    # summarize results
    summaries, state = summarize_results(
        prompt=prompt,
        state=state,
        model=results_summarization,
    )

    # compute answer
    output, state = compute_answer_from_summaries(
        prompt=prompt,
        state=state,
        model=answer_prompt_model,
    )

    return [queries, results, summaries, output, state]


def clear_all():
    state = State.new()
    return (
        "",
        "",
        state.queries.list(),
        state.stack.table(),
        state.summaries.table(),
        state,
    )


def rate_feedback(prompt: str, output: str, state: "State", score: int):
    state.logger.warn(
        {
            "action": "rate_feedback_interaction",
            "prompt": prompt,
            "output": output,
            "score": score,
            "state": state.to_json(),
        }
    )


@dataclass
class State:
    stack: "Stack"
    queries: "Queries"
    summaries: "Summaries"

    def __post_init__(self):
        self.logger = get_logger(__name__)

    @classmethod
    def new(cls) -> "State":
        return State(stack=Stack(), queries=Queries(), summaries=Summaries())

    def to_json(self):
        return {
            "stack": self.stack.to_json(),
            "queries": self.queries.to_json(),
            "summaries": self.summaries.to_json(),
        }

    @classmethod
    def from_json(cls, json):
        return cls(
            stack=Stack.from_json(json["stack"]),
            queries=Queries.from_json(json["queries"]),
            summaries=Summaries.from_json(json["summaries"]),
        )


@cli(
    "app.chat_s2",
    arguments=[
        Argument(
            "-sp",
            "--server-port",
            default=7860,
            help="Port to run the server on",
        ),
        Argument(
            "-sn",
            "--server-name",
            default="localhost",
            help="Server address to run the gradio app at",
        ),
        Argument(
            "-llm",
            "--llm-model",
            default="gpt-3.5-turbo",
            help="Language model to use for the chatbot",
        ),
        Argument(
            "-ok",
            "--openai-key",
            default=os.environ.get("OPENAI_API_KEY"),
            help="OpenAI API key",
        ),
        Argument(
            "-sk",
            "--s2-key",
            default=os.environ.get("S2_API_KEY"),
            help="Semantic Scholar API key",
        ),
        Argument(
            "-gk",
            "--google-custom-search-key",
            default=os.environ.get("GOOGLE_CUSTOM_SEARCH_API_KEY"),
        ),
    ],
    requirements=[
        "requests",
        "openai",
        "jinja2",
        "gradio>=3.23.0",
        "watchtower",
        "logging_json",
        "edlib",
    ],
)
def run_v2_demo(
    server_port: int,
    server_name: str,
    openai_key: str,
    llm_model: str,
    s2_key: str,
    google_custom_search_key: str,
):
    assert openai_key is not None, "OpenAI API key is required"
    assert s2_key is not None, "Semantic Scholar API key is required"
    assert (
        google_custom_search_key is not None
    ), "Google Custom Search API key is required"

    search_endpoint = GoogleSearchEndpoint(api_key=google_custom_search_key)

    with open(CSS_URI, "r") as f:
        css_gradio = f.read()

    with open(LOGO_URI) as f:
        svg_logo = f.read()

    with gr.Blocks(
        title="Talk to Shadow Scholar ",
        css=css_gradio,
        analytics_enabled=False,
    ) as demo:
        gr.HTML(DESCRIPTION.format(svg_logo=svg_logo, model=llm_model))
        state = gr.State(State.new)  # pyright: ignore

        with gr.Row(variant="panel"):
            with gr.Column(scale=3):
                run_all = gr.Button(
                    "Run End-to-End",
                    variant="primary",
                    elem_id="run-end-to-end",
                )
                prompt = gr.Textbox(
                    interactive=True,
                    label="Prompt",
                    lines=5,
                    max_lines=5,
                    placeholder="What would you like to know about?",
                )

            with gr.Column(scale=3):
                clear = gr.Button("Clear All", variant="stop")
                output = gr.Textbox(
                    interactive=False,
                    label="Output",
                    lines=5,
                    placeholder="Provide a prompt to get started",
                )

            with gr.Column(min_width=50):
                gr.HTML('<div id="rate-text">Rate Output</div>')
                with gr.Row(variant="compact"):
                    feedback_positive = gr.Button("👍")
                    feedback_negative = gr.Button("👎")

        gr.Markdown("## 1️⃣ Search Pipeline")
        with gr.Row(variant="panel"):
            with gr.Column(scale=1, min_width=200):
                with gr.Row():
                    with gr.Column(min_width=100):
                        make_queries = gr.Button(
                            "Get Queries",
                            elem_classes="match-height",
                            elem_id="get-queries",
                        )
                    with gr.Column(min_width=50):
                        cnt_query = gr.Number(
                            value=2,
                            label="Max Queries",
                            show_label=False,
                            precision=0,
                        )

                queries = gr.Chatbot(label="Queries")
                query_box = gr.Textbox(
                    lines=1,
                    interactive=True,
                    label="Add Query",
                    placeholder="Type a query and press enter...",
                )

            with gr.Column(scale=2, min_width=300):
                with gr.Row():
                    search = gr.Button(
                        "Search",
                        elem_classes="match-height",
                        elem_id="search",
                    )
                    max_results = gr.Number(
                        value=5,
                        label="Max Results",
                        show_label=False,
                        precision=0,
                    )

                results = gr.DataFrame(
                    headers=["ID", "Paper"],
                    value=[["", ""]],
                    col_count=2,
                    interactive=False,
                    type="array",
                    datatype="markdown",
                    wrap=True,
                    overflow_row_behaviour="paginate",
                )
                paper_box = gr.Textbox(
                    lines=1,
                    interactive=True,
                    label="Add Paper from Semantic Scholar",
                    placeholder="Paste a paper ID or URL...",
                )

        gr.Markdown("## 2️⃣ Rank and Summarize")
        with gr.Row(variant="panel"):
            with gr.Column():
                summarize = gr.Button(
                    value="Summarize Results", elem_id="summarize-results"
                )
                summaries = gr.DataFrame(
                    headers=["Paper", "Summary", "Relevant?"],
                    value=[["", "", ""]],
                    col_count=3,
                    interactive=False,
                    type="array",
                    datatype="markdown",
                    wrap=True,
                    overflow_row_behaviour="paginate",
                )
                compute_answer = gr.Button(
                    value="Compute Answer", elem_id="compute-answer"
                )

        # llm calls
        query_prompt_model = OpenAiPrompt(
            prompt=QUERY_EXTRACTION_TEMPLATE,
            openai_api_key=openai_key,
            model=llm_model,
            temperature=0.7,
        )
        answer_prompt_model = OpenAiPrompt(
            prompt=ANSWER_FORMULATION_TEMPLATE,
            openai_api_key=openai_key,
            model=llm_model,
        )
        results_summarization = OpenAiPrompt(
            prompt=RESULTS_SUMMARIZATION_TEMPLATE,
            openai_api_key=openai_key,
            model=llm_model,
            temperature=0.5,
            max_tokens=128,
        )
        # connect all components here after rendering the UI

        # This allows users to add queries
        query_box.submit(
            add_user_query_to_chatbot,
            [query_box, state],
            [query_box, queries, state],
        )

        paper_box.submit(
            add_user_paper_to_chatbot,
            [paper_box, state],
            [paper_box, results, state],
        )

        # This allows users to automatically generate queries from a prompt
        make_queries.click(
            partial(
                make_queries_from_prompt,
                model=query_prompt_model,
            ),
            [prompt, cnt_query, state],
            [queries, state],
        )

        # this uses google custom search to search for papers
        search.click(
            partial(search_and_table, endpoint=search_endpoint),
            [state, max_results],
            [results, state],
        )

        # this uses gpt to get summaries
        summarize.click(
            partial(summarize_results, model=results_summarization),
            [prompt, state],
            [summaries, state],
        )

        # this gets the final answer
        compute_answer.click(
            partial(compute_answer_from_summaries, model=answer_prompt_model),
            [prompt, state],
            [output, state],
        )

        clear.click(
            clear_all,
            [],
            [prompt, output, queries, results, summaries, state],
        )

        run_all.click(
            partial(
                run_all_fn,
                search_endpoint=search_endpoint,
                query_prompt_model=query_prompt_model,
                answer_prompt_model=answer_prompt_model,
                results_summarization=results_summarization,
            ),
            [prompt, cnt_query, max_results, state],
            [queries, results, summaries, output, state],
        )

        feedback_positive.click(
            partial(rate_feedback, score=1), [prompt, output, state]
        )
        feedback_negative.click(
            partial(rate_feedback, score=-1), [prompt, output, state]
        )

    try:
        demo.launch(
            server_name=server_name,
            server_port=server_port,
            show_api=False,
        )
    except Exception as e:
        gr.close_all()
        raise e



from abc import ABC, abstractmethod, abstractproperty
from enum import IntEnum
from typing import Any, Dict, Literal, Sequence, Union

from shadow_scholar.cli import safe_import

with safe_import():
    import numpy as np
    import torch
    from unqlite import UnQLite


TensorType = Union[torch.Tensor, np.ndarray]


class CacheMode(IntEnum):
    READ_ONLY = 1
    READ_WRITE = 4


class BaseTensorCache(ABC):
    def __init__(
        self,
        *args,
        mode: CacheMode = CacheMode.READ_WRITE,
        invalidate: bool = False,
        **kwargs,
    ) -> None:
        ...

    @abstractproperty
    def signature(self) -> str:
        ...

    @abstractmethod
    def write_many(self, keys: Sequence[str], values: Sequence[TensorType]):
        ...

    def write_one(self, key: str, value: TensorType):
        self.write_many([key], [value])

    @abstractmethod
    def read_many(self, keys: Sequence[str]) -> Sequence[TensorType]:
        ...

    def read_one(self, key: str) -> TensorType:
        return self.read_many([key])[0]



import os
import sys
from logging import INFO, Formatter, Logger, StreamHandler, getLogger
from typing import Optional

import boto3
import watchtower
from logging_json import JSONFormatter


def get_client():
    region_name = os.environ.get("AWS_REGION", "us-west-2")
    return boto3.client("logs", region_name=region_name)


def get_formatter():
    fields = {
        "level_name": "levelname",
        "timestamp": "asctime",
        "modulename": "module",
        "functionname": "funcName",
    }
    return JSONFormatter(fields=fields)


def get_stream_handler(formatter: Formatter):
    stream_handler = StreamHandler(stream=sys.stdout)
    stream_handler.setFormatter(formatter)
    return stream_handler


def get_cloudwatch_handler(formatter: Formatter, client=None):
    cloudwatch_handler = watchtower.CloudWatchLogHandler(
        log_group="shadow_scholar",
        stream_name="chat_s2",
        boto3_client=client,
        send_interval=15,
    )
    cloudwatch_handler.setFormatter(formatter)
    return cloudwatch_handler


def get_logger(
    name: Optional[str] = None,
    disable_logging: str = os.environ.get("DISABLE_ALL_LOGGING", ""),
    disable_cloudwatch: str = os.environ.get("DISABLE_CLOUDWATCH_LOGGING", ""),
) -> Logger:
    name = f"chat_s2.{name}" if name else "chat_s2"

    logger = getLogger(name)
    logger.setLevel(INFO)

    if disable_logging:
        return logger

    formatter = get_formatter()
    stream_handler = get_stream_handler(formatter)
    logger.addHandler(stream_handler)

    if disable_cloudwatch:
        return logger

    try:
        client = get_client()
        cloudwatch_handler = get_cloudwatch_handler(
            formatter=formatter, client=client
        )
        logger.addHandler(cloudwatch_handler)
    except Exception:
        logger.error("Failed to connect to CloudWatch")

    return logger


if __name__ == "__main__":
    logger = get_logger(__name__)
    for msg in sys.argv[1:]:
        logger.info(msg)



import inspect
import json
import shutil
import sys
from argparse import ArgumentParser
from contextlib import contextmanager
from functools import partial
from math import ceil, floor
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from necessary import necessary
from typing_extensions import ParamSpec, TypeAlias

# parameter spec
PS = ParamSpec("PS")

# return type
RT = TypeVar("RT")

# generic type
T = TypeVar("T")

# type alias for a requirement
RequirementType: TypeAlias = Union[str, Tuple[str, str], "NamedRequirement"]


# sentinel value for missing arguments
class _M:
    ...  # noqa: E701


M = _M()  # noqa: E305


class ShadowModuleNotFoundError(ModuleNotFoundError):
    # make easier to track when this error is raised
    ...


class Argument:
    """Holds parameters for argparse.ArgumentParser.add_argument."""

    args: Tuple[str, ...]
    kwargs: Dict[str, Any]

    def __init__(
        self,
        *name_or_flags: str,
        action: Optional[str] = M,  # type: ignore
        nargs: Optional[str] = M,  # type: ignore
        const: Optional[str] = M,  # type: ignore
        default: Optional[T] = M,  # type: ignore
        type: Union[Type[T], Callable[..., T], None] = _M,  # type: ignore
        choices: Optional[List[str]] = M,  # type: ignore
        required: Optional[bool] = M,  # type: ignore
        help: Optional[str] = M,  # type: ignore
        metavar: Optional[str] = M,  # type: ignore
        dest: Optional[str] = M,  # type: ignore
    ):
        self.args = name_or_flags
        self.kwargs = {
            **({"action": action} if action is not M else {}),
            **({"nargs": nargs} if nargs is not M else {}),
            **({"const": const} if const is not M else {}),
            **({"default": default} if default is not M else {}),
            **({"type": type} if type is not _M else {}),
            **({"choices": choices} if choices is not M else {}),
            **({"required": required} if required is not M else {}),
            **({"help": help} if help is not M else {}),
            **({"metavar": metavar} if metavar is not M else {}),
            **({"dest": dest} if dest is not M else {}),
        }


class SpecTuple(NamedTuple):
    name: str
    func: str
    module: str


class NamedRequirement(NamedTuple):
    name: str
    package: str

    @classmethod
    def parse(cls, elem: RequirementType) -> "NamedRequirement":
        if isinstance(elem, cls):
            return elem
        if isinstance(elem, str):
            return cls(elem, elem)
        elif isinstance(elem, tuple):
            return cls(*elem)
        else:
            raise TypeError(f"Expected str or 2-tuple, got {type(elem)}")

    @classmethod
    def from_list(
        cls, elems: List[Union[str, Tuple[str, str]]]
    ) -> List["NamedRequirement"]:
        return [cls.parse(elem) for elem in elems]


class EntryPoint(Generic[PS, RT]):
    def __init__(
        self,
        name: str,
        func: Callable[PS, RT],
        args: List[Argument],
        reqs: List[RequirementType],
    ):
        self.name = name
        self.func = func
        self.args = args

        # if receiving a tuple, the first element is the name of the
        # package and the second is the name of the module to import
        self.reqs = NamedRequirement.from_list(reqs)

        # check that all requirements are met (error raised later if not)
        self.missing_reqs = [
            r.name for r in self.reqs if not necessary(r.package, soft=True)
        ]

    @property
    def spec(self) -> SpecTuple:
        """Returns a triple that identifies this entry point.

        The first argument of the triple is the name of the entry point,
        the second is the name of the function, and the third is the path
        to the function.
        """
        return SpecTuple(self.name, self.func.__name__, self.func.__module__)

    def __call__(self, *args: PS.args, **kwargs: PS.kwargs) -> RT:
        """Run the function."""

        if self.missing_reqs:
            raise ShadowModuleNotFoundError(
                f"Missing requirements: {' '.join(self.missing_reqs)}"
            )
        return self.func(*args, **kwargs)

    def cli(
        self,
        args: Optional[List[str]] = None,
        file_config: Optional[Dict[str, Any]] = None,
    ) -> RT:
        """Run the function from the command line."""

        file_config = file_config or {}

        ap = ArgumentParser(f"shadow-scholar {self.name}")
        for arg in self.args:
            ap.add_argument(*arg.args, **arg.kwargs)

        # parse command line arguments, merge them with json config
        # if one is provided.
        opts, *_ = ap.parse_known_args(args)
        cli_config = vars(opts)

        # only override with file options that are not already set
        # by the command line or that have a value that is equal to defaults
        config = {
            **cli_config,
            **{
                k: v
                for k, v in file_config.items()
                if k not in cli_config or ap.get_default(k) == cli_config[k]
            },
        }

        # actually bind the arguments to the to function args and run the cli
        parsed_args = inspect.signature(self.func).bind(**config).arguments
        return self(**parsed_args)  # pyright: ignore

    @classmethod
    def decorate(
        cls,
        func: Callable[PS, RT],
        name: Optional[str] = None,
        arguments: Optional[List[Argument]] = None,
        requirements: Optional[List[RequirementType]] = None,
        registry: Optional["Registry"] = None,
    ) -> "EntryPoint[PS, RT]":
        """Decorator designed to add function to a registry alongside
        all its arguments and requirements.

        We add the requirement for future parsing, rather than
        building a ArgumentParser here, for two reasons:
        1. We want to be able for all decorated functions to be able to
            use overlapping arguments, and
        2. Creating a single parser here might lead to a lot of unintended
            side effects.

        Args:
            func (Callable): The function to be decorated.
            name (str, optional): Name to use for the function when it is
                called from the command line. If none is provided, the
                function name is used. Defaults to None.
            arguments (List[Argument], optional): A list of Argument objects
                to be passed to the ArgumentParser. The available options
                are the same as those for argparse.ArgumentParser.add_argument.
                Defaults to None.
            requirements (Optional[List[str]], optional): A list of required
                packages for the function to run. Defaults to None.
            registry (Registry, optional): The registry to add the function
                to. If none is provided, the entrypoint is not registered.
        """

        name = name or func.__name__
        arguments = arguments or []
        func_requirements = requirements or []

        # create the entry point by wrapping the function
        entry_point = cls(
            name=name,
            func=func,
            args=arguments,
            reqs=func_requirements,
        )

        # add the function to the registry
        registry.add(entry_point) if registry else None

        return entry_point


class Registry:
    """A registry to hold all the functions decorated with @cli."""

    __instance__: "Registry"
    _registry: Dict[str, "EntryPoint"]

    def __new__(cls):
        """Singleton pattern for the registry."""
        if not hasattr(cls, "__instance__"):
            cls.__instance__ = super(Registry, cls).__new__(cls)
        return cls.__instance__

    def __init__(self) -> None:
        if not hasattr(self, "_registry"):
            self._registry = {}

    def add(self, entry_point: EntryPoint) -> None:
        """Add an entry point to the registry."""
        _, _, module = entry_point.spec

        # we raise an error if the function is already in the registry
        # and the user is not using a different entry point
        if entry_point.name in self._registry and module != "__main__":
            raise KeyError(f"Func {entry_point.name} already in the registry")

        self._registry[entry_point.name] = entry_point

    def cli(
        self,
        name: Optional[str] = None,
        arguments: Optional[List[Argument]] = None,
        requirements: Optional[List[RequirementType]] = None,
    ) -> Callable[[Callable[PS, RT]], "EntryPoint[PS, RT]"]:
        """A decorator to add a function to the registry.

        Args:
            name (str, optional): Name to use for the function
                when it is called from the command line. If none is provided,
                the function name is used. Defaults to None.
            arguments (List[Argument], optional): A list of Argument objects
                to be passed to the ArgumentParser. The available options
                are the same as those for argparse.ArgumentParser.add_argument.
                Defaults to None.
            requirements: A list of requirements for the function. Can
                be either a list of strings (e.g., ["spacy"]), a list that
                includes the minimum versions (e.g., ["spacy>=3.0.0"]), or a
                list of tuples (e.g., [("scikit-learn>=0.24.0", "sklearn"),
                "spacy"]) for cases where the package name is different from
                the import name.
        """
        decorated = partial(
            # need to make types more permissive here because otherwise
            # mypy complains about the partial not being a callable
            cast(Callable[..., EntryPoint[PS, RT]], EntryPoint.decorate),
            name=name,
            arguments=arguments,
            requirements=requirements,
            registry=self,
        )
        return decorated  # pyright: ignore

    def _formatted_entrypoints(self, sep="   ") -> str:
        term_width = shutil.get_terminal_size().columns

        l_col, r_col = zip(
            *(
                (e.spec.name, f"{e.spec.module}.{e.spec.func}")
                for e in sorted(self._registry.values(), key=lambda e: e.name)
            )
        )

        # pad the left column to the longest name, trim if too long
        longest_left = max(len(name) for name in l_col)
        l_col = tuple(
            name.ljust(longest_left)
            if len(name) < term_width
            else (name[: term_width - 3] + "...")
            for name in l_col
        )

        # trim the right column if too long
        max_length_right = term_width - longest_left - len(sep)
        r_col = tuple(
            name
            if len(name) <= max_length_right
            else (name[: max_length_right - 3] + "...")
            for name in r_col
        )
        longest_right = max(len(name) for name in r_col)

        # this is the total width of the entire table
        format_width = longest_left + len(sep) + longest_right

        title = "Available Entrypoints"
        title_pad_width = (format_width - len(title) - 2) / 2

        return (
            "=" * format_width
            + "\n"
            + " " * floor(title_pad_width)
            + "Available Entrypoints"
            + " " * ceil(title_pad_width)
            + "\n"
            + "-" * format_width
            + "\n"
            + "\n".join(f"{lc}   {rc}" for lc, rc in zip(l_col, r_col))
            + "\n"
            + "=" * format_width
        )

    def run(self):
        """Creates a click command group for all registered functions."""
        parser = ArgumentParser("shadow-scholar")
        parser.add_argument(
            "-r",
            "--list-requirements",
            action="store_true",
            help="List the requirements for all entrypoints.",
        )
        parser.add_argument(
            "-l",
            "--list-entrypoints",
            action="store_true",
            help="List the requirements for all entrypoints.",
        )
        parser.add_argument(
            "-c",
            "--config-path",
            help="Path to the config file. Complements cli options.",
        )

        # command line arguments; excluding the first one, which is the
        # name of the script
        args_ = sys.argv[1:]

        # the entrypoint could be any of the arguments that is not an option
        entrypoint_positions = [
            i
            for i in range(len(args_))
            if not (
                args_[i].startswith("-")  # ... is an option
                or args_[i - 1] == "-c"  # ... its the path for config
                or args_[i - 1] == "--config-path"  # ... same as above
            )
        ]
        if len(entrypoint_positions) > 0:
            # separate between the global arguments and the any
            # argument that comes after the mention of the first entrypoint
            pre_args, entry_name, post_args = (
                args_[: (ep := entrypoint_positions[0])],
                args_[ep],
                args_[ep + 1 :],
            )
        else:
            pre_args, entry_name, post_args = args_, None, []

        # get any options for shadow-scholar cli itself
        opts = parser.parse_args(pre_args)

        if opts.list_entrypoints:
            print(self._formatted_entrypoints())
            sys.exit(1)

        entrypoint = self._registry.get(entry_name, None)  # pyright: ignore
        if entrypoint is None:
            msg = (
                "No entrypoint "
                + (f"with name `{entry_name}`." if entry_name else "provided.")
                + " Please run again with one of the following names: "
                + ", ".join(sorted(self._registry))
            )
            print(msg, file=sys.stderr)
            sys.exit(1)

        if opts.list_requirements:
            print(" ".join(r.name for r in entrypoint.reqs))
            sys.exit(0)

        config = {}
        if opts.config_path:
            with open(opts.config_path) as f:
                config = json.load(f)

        try:
            entrypoint.cli(args=post_args, file_config=config)
        except ShadowModuleNotFoundError as e:
            e.msg += (
                f"; run `shadow -r {entry_name}` to list all requirements "
                "for this entrypoint."
            )
            raise e

    def require(
        self, requirements: List[RequirementType]
    ) -> Callable[[Callable[PS, RT]], EntryPoint[PS, RT]]:
        """Specifies a list of requirements for a function.

        Args:
            requirements: A list of requirements for the function. Can
                be either a list of strings (e.g., ["spacy"]), a list that
                includes the minimum versions (e.g., ["spacy>=3.0.0"]), or a
                list of tuples (e.g., [("scikit-learn>=0.24.0", "sklearn"),
                "spacy"]) for cases where the package name is different from
                the import name.
        """

        decorated = partial(
            cast(Callable[..., EntryPoint[PS, RT]], EntryPoint.decorate),
            requirements=requirements,
        )
        return decorated  # type: ignore


@contextmanager
def safe_import():
    """Context manager to safely import a package."""
    try:
        yield
    except (ModuleNotFoundError, ImportError):
        pass


def load_kwargs(path_or_json: str) -> Dict[str, Any]:
    """Load a json file containing a config, or a json string."""

    if Path(path_or_json).exists():
        with open(path_or_json) as f:
            path_or_json = f.read()

    try:
        return json.loads(path_or_json)
    except json.JSONDecodeError as e:
        err = ValueError(f"Could not parse json {path_or_json}.")
        raise err from e


run = Registry().run
cli = Registry().cli
require = Registry().require

__all__ = ["run", "cli", "Argument", "safe_import", "load_kwargs", "require"]



import json
from collections import abc
from pathlib import Path
from typing import Iterable, List, Optional, Tuple, Type, TypeVar, Union

from typing_extensions import TypeAlias

from .reg_utils import CallableRegistry

IdType: TypeAlias = Union[str, int]

dataset_registry: CallableRegistry["Dataset"] = CallableRegistry()


class Container:
    __slots__: Tuple[str, ...]

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def fields(cls) -> Tuple[str, ...]:
        return tuple(getattr(cls, "__annotations__", {}).keys())

    @classmethod
    def _recursive(cls, obj):
        if hasattr(obj, "as_dict"):
            return obj.as_dict()
        if isinstance(obj, abc.Mapping):
            return {k: cls._recursive(v) for k, v in obj.items()}
        elif isinstance(obj, abc.Iterable) and not isinstance(obj, str):
            return [cls._recursive(v) for v in obj]
        else:
            return obj

    def as_dict(self) -> dict:
        return {k: self._recursive(getattr(self, k)) for k in self.fields()}


class Id(Container):
    __slots__ = ("id",)

    id: Tuple[IdType, ...]

    def __init__(self, id: IdType, *sub_ids):
        self.id = (id, *sub_ids)

    def __str__(self):
        return ".".join(str(i) for i in self.id)

    @classmethod
    def parse(cls, s: Union[IdType, "Id"]) -> "Id":
        return Id(s) if isinstance(s, (str, int)) else s

    def as_dict(self) -> IdType:  # type: ignore
        return self.id[0]

    @property
    def head(self) -> IdType:
        return self.id[0]

    def __eq__(self, other) -> bool:
        if not isinstance(other, Id):
            return False
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)


class Query(Container):
    __slots__ = "qid", "text"

    qid: Id
    text: str

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.qid = Id.parse(self.qid)


class Document(Container):
    __slots__ = "did", "text", "label", "score"

    did: Id
    text: str
    label: Optional[int]
    score: Optional[float]
    qid: Optional[Id]

    def __init__(self, **kwargs):
        super().__init__(
            **{"label": None, "score": None, "qid": None, **kwargs}
        )
        self.did = Id.parse(self.did)

    def __hash__(self) -> int:
        return hash((self.did, self.text))

    def new(self, **overrides) -> "Document":
        return type(self)(**{**self.as_dict(), **overrides})

    def judge(self, label: int) -> "Document":
        return self.new(label=label)

    def rank(self, score: float) -> "Document":
        return self.new(score=score)

    def query(self, qid: Id) -> "Document":
        return self.new(qid=qid)


class Qrel(Container):
    __slots__ = "qid", "did", "label"

    qid: Id
    did: Id
    label: int

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.qid = Id.parse(self.qid)
        self.did = Id.parse(self.did)


class Dataset(Container):
    __fields__ = "queries", "documents", "qrels"

    queries: List[Query]
    documents: List[Document]
    qrels: List[Qrel]

    def __init__(self, **kwargs):
        super().__init__(
            **{"queries": [], "documents": [], "qrels": [], **kwargs}
        )

    def iter_docs(self) -> Iterable[Document]:
        return (d for d in self.documents)

    def _iter_qrels(self, query: Query, open_set: bool) -> List[Document]:
        qrels = {
            qrel.did.head: qrel.label
            for qrel in self.qrels
            if qrel.qid == query.qid
        }
        docs = [
            doc.judge(qrels.get(doc.did.head, 0))
            for doc in self.documents
            if (open_set or doc.did.head in qrels)
        ]
        return docs

    def iter_close(self, query: Query) -> List[Document]:
        return self._iter_qrels(query=query, open_set=False)

    def iter_open(self, query: Query) -> List[Document]:
        return self._iter_qrels(query=query, open_set=True)


T = TypeVar("T", bound=Container)


def _read_paths(
    root: Path,
    cls: Type[T],
) -> Iterable[T]:
    all_files = [root]

    while len(all_files) > 0:
        path = all_files.pop()
        if path.is_dir():
            all_files.extend(path.iterdir())
            continue

        with open(path) as f:
            for line in f:
                elem = json.loads(line.strip())
                yield cls(**elem)


@dataset_registry.add("from_files")
def from_files(
    docs_path: Union[str, Path],
    queries_path: Optional[Union[str, Path]] = None,
    qrels_path: Optional[Union[str, Path]] = None,
) -> Dataset:
    dataset = Dataset()

    for doc in _read_paths(root=Path(docs_path), cls=Document):
        dataset.documents.append(doc)

    if queries_path is not None:
        for query in _read_paths(root=Path(queries_path), cls=Query):
            dataset.queries.append(query)

    if qrels_path is not None:
        for qrel in _read_paths(root=Path(qrels_path), cls=Qrel):
            dataset.qrels.append(qrel)

    return dataset



import time
from datetime import datetime
from pathlib import Path
from string import Formatter
from typing import Any, Dict, Optional, Union, cast

from shadow_scholar.cli import Argument, cli, safe_import

with safe_import():
    import boto3
    from botocore.client import BaseClient
    from smashed.utils.io_utils import (
        MultiPath,
        copy_directory,
        remove_directory,
        remove_file,
    )


QUERIES_DIR = Path(__file__).parent / "queries"


def wait_for_athena_query(
    client, execution_id: str, timeout: int = 5, max_wait: int = 3_600
) -> bool:
    state = "RUNNING"

    print(f"Waiting for {execution_id} to complete..", end="", flush=True)

    while max_wait > 0 and state in ["RUNNING", "QUEUED"]:
        response = client.get_query_execution(QueryExecutionId=execution_id)
        if (
            "QueryExecution" in response
            and "Status" in response["QueryExecution"]
            and "State" in response["QueryExecution"]["Status"]
        ):
            state = response["QueryExecution"]["Status"]["State"]
            if state == "SUCCEEDED":
                print(f"\nQuery {execution_id} succeeded!", flush=True)
                return True
            elif state == "FAILED":
                err = response["QueryExecution"]["Status"]["StateChangeReason"]
                raise RuntimeError(f"Query {execution_id} failed: {err}")

        time.sleep(timeout)
        print(".", end="", flush=True)
        max_wait -= timeout

    raise RuntimeError(f"Query {execution_id} timed out")


@cli(
    name="collections.run_athena_query",
    arguments=[
        Argument(
            "-q",
            "--query-path",
            type=str,
            required=True,
            help="Path to the sql query file to run.",
        ),
        Argument(
            "--s3-staging",
            default="s3://ai2-s2-research/temp/",
            type=str,
            help=(
                "S3 bucket for output of Athena query; to be removed after"
                " execution if -o/--output-location is a local directory."
            ),
        ),
        Argument(
            "--output-name",
            default=datetime.now().strftime("%Y-%m-%d-%H-%M-%S"),
            type=str,
            help="Name of output directory; by default, current date/time",
        ),
    ],
    requirements=["boto3", "botocore", "smashed[remote]"],
)
def run_athena_query_and_get_result(
    query_path: str,
    s3_staging: Union[str, "MultiPath"],
    output_location: Union[str, "MultiPath"],
    output_name: str,
    template_args: Optional[Dict[str, Any]] = None,
):
    s3_staging = MultiPath.parse(s3_staging)
    output_location = MultiPath.parse(output_location)
    s3_staging = output_location if output_location.is_s3 else s3_staging
    s3_output_location = s3_staging / output_name
    template_args = template_args or {}

    with open(query_path) as f:
        query_string = f.read()

    formatting_keys = [
        t[1] for t in Formatter().parse(query_string) if t[1] is not None
    ]
    formatting_data = {
        **{k: "" for k in formatting_keys},
        **(template_args or {}),
    }

    formatted_query_string = query_string.format(**formatting_data)

    athena_client = cast(BaseClient, boto3.client("athena"))
    s3_client = cast(BaseClient, boto3.client("s3"))

    response = athena_client.start_query_execution(
        QueryString=formatted_query_string,
        ResultConfiguration=dict(OutputLocation=s3_staging.as_str),
    )
    execution_id = response["QueryExecutionId"]
    wait_for_athena_query(athena_client, response["QueryExecutionId"])

    # remove metadata and execution manifest from temporary S3 bucket
    manifest_loc = s3_staging / f"{execution_id}-manifest.csv"
    remove_file(manifest_loc, client=s3_client)

    metadata_loc = s3_staging / f"{execution_id}.metadata"
    remove_file(metadata_loc, client=s3_client)

    if not output_location.is_s3:
        copy_directory(s3_output_location, output_location)
        remove_directory(s3_output_location, client=s3_client)

    print(f"Dataset written to {output_location}")


@cli(
    name="collections.s2ag.abstracts",
    arguments=[
        Argument(
            "-l",
            "--limit",
            default=10,
            type=int,
            help="Limit number of results; if 0, no limit",
        ),
        Argument(
            "-a",
            "--abstract",
            default=None,
            type=str,
            help="text to search in the abstract",
        ),
        *run_athena_query_and_get_result.args,
    ],
    requirements=run_athena_query_and_get_result.reqs,  # type: ignore
)
def get_s2ag_abstracts(
    database: str,
    release: str,
    limit: int,
    output_location: str,
    s3_staging: str,
    output_name: str,
    query_path: str = str(QUERIES_DIR / "s2ag_abstracts.sql"),
):
    """Get a sample of the S2ORC dataset from Athena."""

    template_args = {
        "limit_clause": f"LIMIT {limit}" if limit > 0 else "",
    }

    run_athena_query_and_get_result(
        query_path=query_path,
        s3_staging=s3_staging,
        output_location=output_location,
        output_name=output_name,
        template_args=template_args,
    )



"""
Code from adapted from https://github.com/paperswithcode/galai/blob/1a1e4486fd38c85c6b66b3c0b31ce24b0afd7613/galai/model.py

Adapted by @soldni

"""  # noqa: E501

import warnings
from typing import Any, Dict, List, Literal, Optional, Tuple, Union, cast

from shadow_scholar.cli import safe_import

# from .convert_to_int8 import replace_linear_with_int8
from .galai_utils import escape_custom_split_sequence

with safe_import():
    import psutil
    import torch
    from accelerate import infer_auto_device_map, init_empty_weights
    from transformers import (
        AutoModelForCausalLM,
        AutoTokenizer,
        OPTConfig,
        OPTForCausalLM,
        PreTrainedTokenizerBase,
    )


class Model:
    """
    Model class holding the GALACTICA models. We configure a class to
    encapsulate the HuggingFace model, the tokenizer, and the specific
    tokenization logic for GALACTICA. For low-level access, we recommend
    using the standard HuggingFace API.
    """

    model: "OPTForCausalLM"
    tokenizer: "PreTrainedTokenizerBase"

    def __init__(
        self,
        name: str,
        precision: Literal["full", "mixed", "byte"] = "full",
        tensor_parallel: bool = False,
        leftover_space: float = 0.3,
    ):
        """
        Initializes a new model

        Args:
            name (str): The name of the model to load. One of
                `mini`, `base`, `standard`, `large`, `huge`.
            precision (str): The precision to use for the model. One of
                `full` or `mixed`. `full` uses 32-bit floating point
                numbers for the model weights and `mixed` uses 16-bit
                floating point numbers (bfloat16).
            tensor_parallel (bool): Whether to use tensor parallelism
                to load the model. This is only supported on NVIDIA
                GPUs at requires 2 or more GPUs.
            leftover_space (float): The amount of space to leave free
                on the GPU when loading the model. This is useful if
                you are expecting to run long sequences on the model.
        """

        self.name, self.dtype = self.variants[name]
        if precision == "mixed":
            self.dtype = torch.bfloat16
        elif precision == "byte":
            self.dtype = torch.float16

        self.convert_to_int8 = precision == "byte"
        self.is_loaded = False
        self.num_gpus = (
            torch.cuda.device_count()
            if tensor_parallel
            else (1 if torch.cuda.is_available() else 0)
        )
        self.tensor_parallel = tensor_parallel
        self.max_input_length = 2048
        self._master_port: Union[int, None] = None
        self.leftover_space = leftover_space

        self._set_tokenizer(self.name)
        self._load_checkpoint(self.name)

    @property
    def variants(cls) -> Dict[str, Tuple[str, "torch.dtype"]]:
        return {
            "mini": ("facebook/galactica-125m", torch.float32),
            "base": ("facebook/galactica-1.3b", torch.float32),
            "standard": ("facebook/galactica-6.7b", torch.float32),
            "large": ("facebook/galactica-30b", torch.float32),
            "huge": ("facebook/galactica-120b", torch.float16),
        }

    def __str__(self):
        return f"model `{self.name}` with type `{self.dtype}`"

    def _load_checkpoint(self, _: str):
        if self.tensor_parallel:
            config = OPTConfig.from_pretrained(self.name)
            with init_empty_weights():  # pyright: ignore
                empty_model = AutoModelForCausalLM.from_config(config)

            from ._old import _gpu_mem

            device_map = infer_auto_device_map(
                empty_model,
                max_memory={
                    gid: f"{mem * (1 - self.leftover_space):.0f}MB"
                    for gid, mem in _gpu_mem().items()
                },
                no_split_module_classes=[
                    "OPTDecoderLayer",
                    "Linear" "Embedding",
                    "OPTLearnedPositionalEmbedding" "LayerNorm",
                ],
                dtype=self.dtype,
            )

            # HACK: we actually want to have "lm_head" on the first GPU
            #       because its weights are tied to the embedding layer
            #       and we want to keep them on the same device.
            if "lm_head" in device_map:
                device_map["lm_head"] = 0
            assert (
                device_map.get(
                    "model.decoder.embed_tokens",
                    0,
                )
                == device_map.get(
                    "model.decoder.embed_positions",
                    0,
                )
                == device_map.get("model.decoder.final_layer_norm", 0)
                == device_map.get("lm_head", 0)
            ), "Embedding layers and lm_head should be on the same device."

            self.model = OPTForCausalLM.from_pretrained(
                pretrained_model_name_or_path=self.name,
                device_map=device_map,
                offload_folder="offload",
                offload_state_dict=True,
                torch_dtype=self.dtype,
            ).eval()  # pyright: ignore

        else:
            device = "cuda" if torch.cuda.is_available() else "cpu"
            self.model = (
                OPTForCausalLM.from_pretrained(
                    pretrained_model_name_or_path=self.name,
                    torch_dtype=self.dtype,
                )
                .to(device)  # pyright: ignore
                .eval()
            )

        if self.convert_to_int8:
            raise NotImplementedError(
                "Conversion to int8 is not supported yet. "
            )
            # replace_linear_with_int8(
            #     self.model,
            #     do_not_convert=[
            #         "lm_head",
            #         # 'model.decoder.embed_tokens',
            #         # 'model.decoder.embed_positions'
            #     ],
            # )

    def _buggy_load_checkpoint(self, checkpoint_path: str):
        """
        Loads the checkpoint for the model

        Parameters
        ----------
        checkpoint_path : str
            Path for the checkpoint (str)
        """

        # query available memory size of the GPUs we want to use.
        # If tensor_parallel is True, we just load the model's weights to RAM,
        # as it needs to be sliced by parallelformers before loading to VRAM.
        device_map = None
        max_memory: Dict[Union[str, int], int] = {}
        if self.num_gpus > 0 and not self.tensor_parallel:
            # based on https://github.com/huggingface/accelerate/blob/5315290b55ea9babd95a281a27c51d87b89d7c85/src/accelerate/utils/modeling.py#L274    # noqa: E501
            for i in range(self.num_gpus):
                _ = torch.tensor([0], device=i)
            for i in range(self.num_gpus):
                max_memory[i] = torch.cuda.mem_get_info(i)[0]
            device_map = "auto"
        max_memory["cpu"] = psutil.virtual_memory().available

        self.model = OPTForCausalLM.from_pretrained(  # pyright: ignore
            checkpoint_path,
            torch_dtype=self.dtype,
            low_cpu_mem_usage=True,
            device_map=device_map,
            max_memory=max_memory,
        )
        self.model.eval()  # pyright: ignore

        if self.tensor_parallel:
            self._parallelize()

    def _parallelize(self) -> None:
        """
        Parallelize the model for a tensor-parallel multi-GPU inference.
        """

        if self.num_gpus < 2:
            raise RuntimeError(
                "At least two GPUs are required to parallelize the model."
            )

        self._master_port = 13000 + (id(self.model) % 32749)

        custom_policies = None
        if (
            self.model.config.model_type == "opt"
            and not self.model.config.enable_bias
        ):
            from .galai_parallel_policy import OPTDecoderLayerPolicyNoBias

            custom_policies = [OPTDecoderLayerPolicyNoBias]

        from parallelformers import parallelize

        parallelize(
            self.model,
            num_gpus=self.num_gpus,
            fp16=self.dtype == torch.float16,
            master_port=self._master_port,
            custom_policies=custom_policies,
        )

    def _set_tokenizer(self, tokenizer_path: str):
        """
        Configures the tokenizer for the model

        Parameters
        ----------
        tokenizer_path : str
            Path for the tokenizer (str)
        """
        tokenizer = AutoTokenizer.from_pretrained(tokenizer_path)

        # setup padding
        tokenizer.pad_token_id = 1
        tokenizer.pad_token = "<pad>"
        tokenizer.padding_side = "left"

        # setup truncation
        tokenizer.truncation_side = "left"

        # setup special tokens
        tokenizer.bos_token_id = 0
        tokenizer.bos_token = "<s>"

        tokenizer.eos_token_id = 2
        tokenizer.eos_token = "</s>"

        tokenizer.unk_token = "<unk>"
        tokenizer.unk_token_id = 3

        self.tokenizer = tokenizer

    def _tokenize(
        self, input_text: List[str], new_doc: bool
    ) -> "torch.LongTensor":
        """
        Apply custom preprocessing to input texts and tokenize them.

        Args:
            input_text (List[str]): Texts to be tokenized
            new_doc (bool): If True, prepends the end-of-document (</s>)
                token to each sequence.
        """
        texts = []
        for text in input_text:
            text = escape_custom_split_sequence(text)
            if not text:
                warnings.warn(
                    "Found an empty input text. Changing to "
                    + "end-of-document token instead.",
                    UserWarning,
                )
                text = self.tokenizer.eos_token
            texts.append(text)

        if new_doc:
            pad_token = self.tokenizer.pad_token
            texts = [pad_token + t for t in texts]

        encoded = self.tokenizer(
            texts,
            padding="longest",
            max_length=self.max_input_length,
            truncation=True,
            pad_to_multiple_of=8,
        )
        context_tokens = encoded["input_ids"]
        input_v = cast(
            torch.LongTensor,
            torch.LongTensor(context_tokens).to(self.model.device),
        )

        if new_doc:
            loc = input_v[:, 0] == self.tokenizer.pad_token_id
            input_v[loc, 0] = cast(int, self.tokenizer.eos_token_id)
        return input_v

    def generate(
        self,
        input_text: Union[str, List[str]],
        max_new_tokens: int = 128,
        new_doc: bool = False,
        top_p: Optional[float] = None,
        top_k: Optional[float] = None,
        penalty_alpha: Optional[float] = None,
        num_beams: int = 1,
        num_return_sequences: int = 1,
        return_full_text: bool = True,
        extra_options: Optional[Dict[str, Any]] = None,
    ) -> Union[str, List[str], List[List[str]]]:
        """
        Generates text using the model

        Args:
            input_text (Union[str, List[str]]): Input context for the model
                to use for its generation, e.g. "Attention Is All You Need
                [START_REF]"
            max_length (int, optional): Maximum length in tokens of the
                generated text (including prompt). Only one of max_length
                and max_new_tokens should be specified. If neither is set,
                then max_new_tokens is set to 60.
            max_new_tokens (int, optional): Maximum length in tokens of the
                generated text (excluding prompt). Only one of max_length
                and max_new_tokens should be specified. If neither is set,
                then max_new_tokens is set to 60.
            new_doc (bool): If True, treats generation a new document,
                otherwise assumes generation could be anywhere within
                document. Use new_doc=True if you are generating documents,
                e.g. # Schwarzschild Radius, # Transformer (machine
                learning), Title: Transformers, A Survey. For general
                generation, set new_doc=False.
            top_p (float, optional): If set, the model will use nucleus
                sampling with this probability. This means that the model
                will sample from the most likely tokens until the sum of
                their probabilities exceeds this value. This is useful for
                generating more diverse outputs. Default is None, which
                means no nucleus sampling is used.
            top_k (int, optional): If set, the model will use top-k
                sampling with this value. This means that the model will
                sample from the top-k most likely tokens. This is useful
                for generating more diverse outputs. Default is None, which
                means no top-k sampling is used.
            penalty_alpha (float, optional): If set, the model will use
                penalty-based sampling with this value. This means that the
                model will sample from the most likely tokens, but will
                penalize tokens that have already been generated. This is
                useful for generating more diverse outputs. Default is
                None, which means no penalty-based sampling is used.
            num_beams (int, optional): If set, the model will use beam
                search with this value. This means that the model will
                generate this many sequences at once, and then return the
                best one. This is useful for generating more diverse
                outputs. Default is 1, which means no beam search is used.
            num_return_sequences (int, optional): If set, the model will
                return this many sequences. Default is 1, which means only
                one sequence is returned.
            return_full_text (bool, optional): If True, the model will
                return the full text of the generated sequence, including
                the prompt. If False, the model will return only the
                generated sequence. Default is True.
        """

        with torch.inference_mode():
            texts = [input_text] if isinstance(input_text, str) else input_text
            input_v = self._tokenize(texts, new_doc)
            options = {}
            if penalty_alpha is not None:
                options["penalty_alpha"] = float(penalty_alpha)
                options["top_k"] = int(top_k) if top_k is not None else 0
            elif top_p is not None:
                options["do_sample"] = True
                options["top_p"] = float(top_p)
            elif top_k is not None:
                options["do_sample"] = True
                options["top_k"] = int(top_k)

            if extra_options:
                options.update(extra_options)

            out = self.model.generate(
                input_v,
                max_new_tokens=int(max_new_tokens),
                return_dict_in_generate=True,
                output_hidden_states=False,
                num_beams=int(num_beams),
                num_return_sequences=int(num_return_sequences),
                **options,
            )

            out_tokens = out["sequences"]  # pyright: ignore
            if not return_full_text:
                out_tokens = out_tokens[:, input_v.shape[1] :]
            # we keep special tokens such as [START_REF] or <work>
            decoded = self.tokenizer.batch_decode(
                out_tokens,
                skip_special_tokens=False,
                clean_up_tokenization_spaces=False,
            )
            # so we manually remove </s> and <pad>
            decoded = [
                text.replace(self.tokenizer.eos_token, "").replace(
                    self.tokenizer.pad_token, ""
                )
                for text in decoded
            ]

            if num_return_sequences == 1:
                return decoded[0] if isinstance(input_text, str) else decoded
            if isinstance(input_text, str):
                return decoded
            else:
                return [
                    decoded[
                        num_return_sequences
                        * i : num_return_sequences
                        * (i + 1)
                    ]
                    for i in range(len(texts))
                ]



from pathlib import Path

from shadow_scholar.cli import Argument, cli, safe_import

with safe_import():
    import gradio as gr

    # from fastapi import FastAPI
    # import uvicorn


S2R_URL = "https://www.semanticscholar.org/research/research-team"
SWS_URL = "https://github.com/allenai/shadow-scholar"
CSS_URI = Path(__file__).parent / "res" / "style.css"
IMG_URI = Path(__file__).parent / "res" / "logo.png"


@cli(
    "app.hello_world",
    arguments=[
        Argument("--host", default="localhost", help="Host to bind to"),
        Argument("--port", default=7860, type=int, help="Port to bind to"),
    ],
    requirements=["gradio"],
)
def run_hello_world(
    host: str,
    port: int,
):
    with open(CSS_URI, "r") as f:
        css = f.read()

    with gr.Blocks(title="Shadow Scholar", css=css) as app:
        gr.Markdown("# 🕶️ 🎓 Shadow Scholar 🎓 🕶️", elem_id="center")
        gr.Image(
            str(IMG_URI),
            elem_id="logo",
            shape=(400, 400),
            show_label=False,
            interactive=False,
        )
        gr.Markdown(
            "Shadow Scholar is a collection of tools and applications "
            f"from the [S2 Research Team]({S2R_URL}). To learn more about "
            f"this project, please visit our [GitHub repository]({SWS_URL}). "
            f"To install shadow scholar, run `pip install shadow-scholar`.",
            elem_id="center",
        )

    try:
        app.launch(
            server_name=host, server_port=port, show_api=False, inline=True
        )
    except Exception as e:
        app.close()
        raise e



"""

@lucas, @kylel

"""

import os
from logging import Logger
from time import time
from typing import Iterable, Optional

from shadow_scholar.cli import require, safe_import

with safe_import():
    import requests

S2_API_KEY = os.environ.get("S2_API_KEY", "")

API_FIELDS = [
    "title",
    "abstract",
    "venue",
    "fieldsOfStudy",
    "authors",
    "isOpenAccess",
    "year",
    "corpusId",
]


@require(["requests"])
def get_s2_metadata(
    paper_ids: Iterable[str],
    s2_api_key: str = S2_API_KEY,
    logger: Optional[Logger] = None,
):
    start = time()

    url = (
        "https://api.semanticscholar.org/graph/v1/paper/batch?"
        f"fields={','.join(API_FIELDS)}"
    )
    header = {"x-api-key": s2_api_key}

    data = {"ids": list(paper_ids)}
    response = requests.post(url, headers=header, json=data)
    results = response.json()

    delta = time() - start
    logger.info(
        {
            "url": url,
            "data": data,
            "elapsed": delta,
            "response": response,
            "action": "get_s2_metadata",
        }
    ) if logger else None

    return results



import html
import math
import re
from dataclasses import dataclass
from typing import List

__all__ = [
    "escape_custom_split_sequence",
    "ModelInfo",
]


# we split individual characters inside special tokens like [START_DNA]
CUSTOM_SEQ_RE = re.compile(
    r"(\[START_(DNA|SMILES|I_SMILES|AMINO)])(.*?)(\[END_\2])"
)

# token added to implement a custom sequence tokenization. This token is
# added at corpus cleaning step and removed in pretokenization. The digits
# are added to increase the chance that they do not occur in the corpus.
# The digits are escaped so that the token does not appear literally in the
# source code in case we ever include it in the training data.
SPLIT_MARKER = f"SPL{1}T-TH{1}S-Pl3A5E"


def _insert_split_marker(m: re.Match):
    """
    Applies split marker based on a regex match of special tokens such as
    [START_DNA].

    Parameters
    ----------
    n : str
        Input text to split

    Returns
    ----------
    str - the text with the split token added
    """
    start_token, _, sequence, end_token = m.groups()
    sequence = re.sub(r"(.)", rf"{SPLIT_MARKER}\1", sequence, flags=re.DOTALL)
    return f"{start_token}{sequence}{SPLIT_MARKER}{end_token}"


def escape_custom_split_sequence(text):
    """
    Applies custom splitting to the text for GALILEO's tokenization

    Parameters
    ----------
    text : str
        Input text to split

    Returns
    ----------
    str - the text with the split token added
    """
    return CUSTOM_SEQ_RE.sub(_insert_split_marker, text)


REFERENCE_RE = re.compile(r"\[START_REF\](.*?)\[END_REF\]", flags=re.DOTALL)


def extract_references_from_text(text: str) -> List[str]:
    return [cit.strip() for cit in REFERENCE_RE.findall(text)]


@dataclass
class ModelInfo:
    name: str
    num_layers: int
    num_heads: int
    head_size: int = 128
    vocab_size: int = 50000
    max_positions: int = 2048

    @property
    def hidden_dimension(self) -> int:
        return self.head_size * self.num_heads

    @property
    def parameters(self) -> int:
        layer_norm_elementwise_affine = True
        enable_bias = True
        h_dim = self.hidden_dimension
        bias = h_dim if enable_bias else 0
        embed_tokens_size = self.vocab_size * h_dim
        embed_positions_size = (self.max_positions + 2) * h_dim
        layer_norm_size = 2 * h_dim if layer_norm_elementwise_affine else 0
        # 4 = k_proj+v_proj+q_proj+out_proj
        self_attn_size = 4 * (h_dim * h_dim + bias)
        ffn_dim = 4 * h_dim
        fc_size = 2 * h_dim * ffn_dim + 5 * bias  # 2 = fc1 + fc2
        decoder_layer_size = self_attn_size + fc_size + 2 * layer_norm_size
        decoder_size = (
            self.num_layers * decoder_layer_size
            + layer_norm_size
            + embed_tokens_size
            + embed_positions_size
        )

        return decoder_size

    @property
    def disk_size(self) -> int:
        """Approximate dist size in bytes of checkpoints files"""
        return self.parameters * 2

    def weights_size(self, dtype="float16") -> int:
        """Approximate total size of model weights in memory"""
        element_size = 2 if dtype == "float16" else 4
        return self.parameters * element_size

    def memory_per_token(self, dtype="float16") -> int:
        """Approximate memory size required to store intermediate activations
        and cached outputs"""
        element_size = 2 if dtype == "float16" else 4
        return (
            2
            * self.num_layers
            * self.num_heads
            * self.head_size
            * element_size
        )

    @staticmethod
    def by_name(name: str) -> "ModelInfo":
        return _MODEL_INFO_BY_NAME[name]

    @staticmethod
    def all() -> List["ModelInfo"]:
        return _MODEL_INFO


def _humanize(parameters):
    scale = min(int(math.log10(parameters)) // 3, 4)
    suffix = " KMBT"[scale]

    return f"{parameters / math.pow(10, 3 * scale):.1f} {suffix}".rstrip()


class ModelInfoList(list):
    def _repr_html_(self):
        if not self:
            return ""
        columns = {
            "Name": lambda m: f"<strong>{html.escape(m.name)}</strong>",
            "Parameters": lambda m: _humanize(m.parameters),
            "Layers": lambda m: str(m.num_layers),
            "Heads": lambda m: str(m.num_heads),
            "Head Size": lambda m: str(m.head_size),
            "Vocabulary Size": lambda m: str(m.vocab_size),
            "Context Size": lambda m: str(m.max_positions),
        }
        output = ["<table><thead><tr>"]
        for col in columns:
            output.append(f"<th>{col}</th>")
        output.append("</tr></thead><tbody>")
        for mi in self:
            output.append("<tr>")
            for extractor in columns.values():
                output.append(f"<td>{extractor(mi)}</td>")
            output.append("</tr>")
        output.append("</tbody></table>")
        return "".join(output)


_MODEL_INFO = ModelInfoList(
    [
        ModelInfo("mini", num_layers=12, num_heads=12, head_size=64),
        ModelInfo("base", num_layers=24, num_heads=32, head_size=64),
        ModelInfo("standard", num_layers=32, num_heads=32, head_size=128),
        ModelInfo("large", num_layers=48, num_heads=56, head_size=128),
        ModelInfo("huge", num_layers=96, num_heads=80, head_size=128),
    ]
)

_MODEL_INFO_BY_NAME = {model.name: model for model in _MODEL_INFO}



import unittest


class TestImports(unittest.TestCase):
    def test_imports(self):
        try:
            import shadow_scholar  # noqa: F401 type: ignore
        except ImportError:
            self.fail(
                "Could not import shadow_scholar; probably dependencies"
                " are not accurately marked as optional."
            )



from shadow_scholar.cli import cli


@cli(name="hello-world")
def hello_world():
    msg = (
        "You have successfully installed shadow-scholar! If you want to see "
        "what other entrypoints are available, run `shadow -l`. To list all "
        "options for a given entrypoint, run `shadow <entrypoint> -h`, or "
        "`shadow <entrypoint> -r` to list its required packages. To check "
        "out how shadow-scholar responds when a package is missing, run "
        "`shadow angry-world`."
    )
    print(msg)


@cli(name="angry-world", requirements=["a-python-package-that-does-not-exist"])
def angry_world():
    ...



import os
import re
from dataclasses import dataclass, field
from logging import Logger
from time import time
from typing import Dict, Iterable, List, Optional, Union

import edlib
import requests

API_FIELDS = [
    "title",
    "abstract",
    "venue",
    "fieldsOfStudy",
    "authors",
    "isOpenAccess",
    "year",
    "corpusId",
    "externalIds",
]

S2_API_KEY = os.environ.get("S2_API_KEY", "")


def clean_html(raw_html: Optional[str] = None) -> str:
    if not (raw_html := (raw_html or "").strip()):
        return raw_html

    return re.sub(
        r"<.*?>|&([a-z0-9]+|#[0-9]{1,6}|#x[0-9a-f]{1,6});", "", raw_html
    )


def get_s2_metadata_single(
    paper: "Paper",
    s2_api_key: str = S2_API_KEY,
    logger: Optional[Logger] = None,
):
    url = (
        "https://api.semanticscholar.org/graph/v1/paper/"
        f"{paper.id}?fields={','.join(API_FIELDS)}"
    )
    header = {"x-api-key": s2_api_key}
    response = requests.get(url, headers=header)
    paper.metadata.update(response.json())


def get_s2_metadata(
    papers: Iterable["Paper"],
    s2_api_key: str = S2_API_KEY,
    logger: Optional[Logger] = None,
):
    start = time()

    url = (
        "https://api.semanticscholar.org/graph/v1/paper/batch?"
        f"fields={','.join(API_FIELDS)}"
    )
    header = {"x-api-key": s2_api_key}

    papers_with_missing_metadata = [p for p in papers if p.missing]

    if not papers_with_missing_metadata:
        # nothing new to fetch metadata for
        return

    data = {"ids": [p.id for p in papers_with_missing_metadata]}
    response = requests.post(url, headers=header, json=data)
    results = response.json()

    if (
        isinstance(results, dict)
        and results.get("message") == "Internal Server Error"
    ):
        for paper in papers_with_missing_metadata:
            get_s2_metadata_single(paper, s2_api_key, logger)
        return

    # responses might not in order of request, so we need to match them up;
    # we do this by calculating similarity
    for paper in papers_with_missing_metadata:
        distance: List[int] = []
        for result in results:
            if (
                paper.id == result.get("paperId", None)
                or paper.id == result.get("corpusId", None)
                or str(paper.id).split(":", 1)[-1]
                in result.get("externalIds", {}).values()
            ):
                distance.append(0)
                break

            elif "title" in result and "title" in paper.metadata:
                distance.append(
                    edlib.align(paper.metadata["title"], result["title"])[
                        "editDistance"
                    ]
                )
            else:
                distance.append(1000)

        if len(distance) > 0 and (best_guess := min(distance)) < 1000:
            location = distance.index(best_guess)
            metadata = results.pop(location)
            paper.metadata.update(metadata)

    delta = time() - start
    logger.info(
        {
            "url": url,
            "data": data,
            "elapsed": delta,
            "response": results,
            "action": "get_s2_metadata",
        }
    ) if logger else None


@dataclass
class Paper:
    id: str
    metadata: dict = field(default_factory=dict)

    def __str__(self) -> str:
        return f"Paper({self.id})"

    @classmethod
    def id_from_s2_url(cls, url: str) -> Union[str, None]:
        is_valid_url = re.match(
            r"https://(www.)?semanticscholar.org/paper.*?/([a-f0-9]+)", url
        )
        if not is_valid_url:
            return None

        sha1 = is_valid_url.group(2)
        return sha1

    @classmethod
    def id_from_arxiv_url(cls, url: str) -> Union[str, None]:
        is_valid_url = re.match(r"https://(www.)?arxiv.org/abs/(.+)", url)
        if not is_valid_url:
            return None

        arxiv_id = is_valid_url.group(2)
        return f"arxiv:{arxiv_id}"

    @classmethod
    def from_url(cls, url: str, **attrs) -> Optional["Paper"]:
        id_ = cls.id_from_s2_url(url) or cls.id_from_arxiv_url(url)
        return cls(id_, metadata=attrs) if id_ else None

    @property
    def short_id(self) -> Optional[int]:
        if short_id := self.metadata.get("short_id", None):
            short_id = int(short_id)
        return short_id

    @property
    def score(self) -> float:
        return self.metadata.get("score", -1.0)

    @property
    def missing(self) -> bool:
        return self.metadata.get("corpusId", None) is None

    @property
    def authors(self) -> List[str]:
        return [author["name"] for author in self.metadata.get("authors", [])]

    @property
    def url(self) -> str:
        return f"https://api.semanticscholar.org/{self.id}"

    @property
    def year(self) -> Union[int, None]:
        if (year := self.metadata.get("year")) is not None:
            year = int(year)
        return year

    @property
    def title(self) -> str:
        return clean_html(self.metadata.get("title", ""))

    @property
    def abstract(self) -> str:
        return clean_html(self.metadata.get("abstract", ""))

    @property
    def is_open_access(self) -> bool:
        return self.metadata.get("isOpenAccess", False)

    @property
    def venue(self) -> str:
        return self.metadata.get("venue", "")

    @property
    def corpus_id(self) -> str:
        return self.metadata.get("corpusId", "")

    def _truncate(self, text: str, cols: int = 70) -> str:
        if len(text) > (cols - 3):
            return text[:cols] + "..."
        return text

    def html_format_title(self) -> str:
        return (
            f'<a href="{self.url}" target="_blank">'
            f"<b>{self._truncate(self.title)}</b></a> ({self.year})"
        )

    def html_format_paper(self) -> str:
        title = self.html_format_title()
        authors = f'<p><i>{self._truncate(", ".join(self.authors))}</i></p>'
        abstract = f'<p>{self._truncate(self.abstract or "")}</p>'
        return "\n".join([title, authors, abstract])

    def html_format_id(self) -> str:
        short_id = (
            f'<span class="span-major">[{self.short_id}]</span>'
            if self.short_id
            else ""
        )

        span_class = "span-minor" if short_id else "span-minor"
        title_id = (
            f'<span class="{span_class}">'
            f'<a href="{self.url}" target="_blank">{self.corpus_id}</a>'
            "</span>"
        )
        score = f"<i>{self.score:.2f}</i>"
        return (
            '<div id="id-score-table-container">'
            f"<p>{short_id}</p><p>{title_id}</p>"
            f'<p id="score-para">{score}</p>'
            "</div>"
        )

    def to_json(self):
        return {
            "id": self.id,
            "metadata": self.metadata,
            "short_id": self.short_id,
        }

    @classmethod
    def from_json(cls, data):
        return cls(**data)


@dataclass
class Stack:
    papers: Dict[str, Paper] = field(default_factory=dict)
    s2_api_key: str = S2_API_KEY

    def _compute_short_ids(self):
        for i, p in enumerate(self.papers.values(), start=1):
            if p.short_id is None:
                p.metadata["short_id"] = i

    def append(self, paper: Paper):
        if paper.id not in self.papers:
            self.papers[paper.id] = paper
            self._compute_short_ids()

    def extend(self, papers: Iterable[Paper]):
        for paper in papers:
            self.append(paper)

    def fetch(self):
        get_s2_metadata(self.papers.values(), s2_api_key=self.s2_api_key)
        for pid in list(self.papers):
            # remove papers that are not found
            if self.papers[pid].missing:
                del self.papers[pid]
        self._compute_short_ids()

    @property
    def sorted(self) -> List[Paper]:
        return sorted(self.papers.values(), key=lambda p: -p.score)

    def __len__(self):
        return len(self.papers)

    def __getitem__(self, key):
        return self.papers[key]

    def __iter__(self):
        return iter(self.papers)

    def __str__(self) -> str:
        return (
            f"Stack_{len(self.papers)}(\n"
            + "\n".join(f"\t{paper}," for paper in self.papers)
            + "\n)"
        )

    def table(self) -> List[List[str]]:
        if len(self) == 0:
            # we need to return something otherwise UI will break
            return [["", ""]]

        return [
            [p.html_format_id(), p.html_format_paper()]
            for p in sorted(
                self.sorted, key=lambda p: (p.short_id or -p.score)
            )
        ]

    def to_json(self):
        return [p.to_json() for p in self.papers.values()]

    @classmethod
    def from_json(cls, data):
        return cls(papers={p["id"]: Paper.from_json(p) for p in data})



import re
from enum import Enum
from typing import Any, Generic, List, Sequence, TypeVar, Union

from cachetools import LRUCache

from shadow_scholar.cli import safe_import

from .datasets import Document
from .reg_utils import CallableRegistry

with safe_import():
    # dependencies for models
    import numpy as np
    import spacy
    import spacy.tokens
    import torch
    from sentence_transformers import SentenceTransformer, models
    from sklearn.feature_extraction.text import (
        CountVectorizer,
        TfidfTransformer,
    )
    from sklearn.pipeline import Pipeline
    from unidecode import unidecode


ranker_registry: CallableRegistry["BaseRanker"] = CallableRegistry()

OutputType = TypeVar("OutputType", "torch.Tensor", "np.ndarray")


class BaseRanker(Generic[OutputType]):
    """Abstract base class for scorers.

    A scorer is a class that can score a list of slices; it can do
    so by either using some model, or some other (un)supervised
    method."""

    lru_cache: Union[LRUCache, None]

    def __init__(self, cache_size: int = 10_000):
        self.lru_cache = (
            LRUCache(maxsize=cache_size) if cache_size > 0 else None
        )

    def warmup(self, docs: Sequence[Document]):
        """Pre-compute the encodings for the given documents"""
        pass

    def _encode(self, batch: Sequence[Document]) -> OutputType:
        """Transform tokens in the query and target spans into a numerical
        representation that can be scored by a model"""
        raise NotImplementedError()

    def encode(
        self, docs: Sequence[Document], **encode_kwargs: Any
    ) -> List[OutputType]:
        batch: List[Document] = []
        for doc in docs:
            if self.lru_cache is None or doc not in self.lru_cache:
                batch.append(doc)

        encoded = self._encode(batch, **encode_kwargs) if batch else []
        cache = self.lru_cache or {}
        for i in range(len(batch)):
            cache[batch[i]] = encoded[i]  # pyright: ignore

        return [cache[doc] for doc in docs]

    def _score(
        self, query: str, docs: Sequence[Document]
    ) -> Sequence[Document]:
        """Score a list of slices based on a query"""
        raise NotImplementedError()

    def _sort_key_fn(self, doc: Document) -> float:
        if doc.score is None:
            raise ValueError("Score is None")
        return -doc.score

    def score(self, query: str, docs: Sequence[Document]) -> List[Document]:
        """Score a list of slices based on a query"""
        return sorted(
            [s for s in self._score(query, docs) if s.score],
            key=self._sort_key_fn,
        )


@ranker_registry.add("tfidf")
class TfIdfRanker(BaseRanker["np.ndarray"]):
    def __init__(
        self,
        cache_size: int = 1024,
        lower: bool = True,
        lemma: bool = True,
        stop: bool = False,
        alpha: bool = True,
        ascii: bool = True,
        spacy_model: str = "en_core_web_sm",
    ) -> None:
        super().__init__(cache_size=cache_size)

        self.lower = lower
        self.lemma = lemma
        self.stop = stop
        self.alpha_fn = (
            lambda x: re.match(r"^[a-zA-Z0-9]+$", x) is not None
            if alpha
            else lambda x: True
        )
        self.ascii_fn = unidecode if ascii else lambda x: x
        self.nlp = spacy.load(spacy_model)

    def _word(self, token: "spacy.tokens.token.Token") -> Union[str, None]:
        if self.stop and token.is_stop:
            return None

        text = token.lemma_ if self.lemma else token.text
        text = text.lower() if self.lower else text
        text = self.ascii_fn(text)

        if not self.alpha_fn(text):
            return None

        return text

    def warmup(self, docs: Sequence[Document]):
        content = [
            [
                w
                for w in (self._word(t) for t in self.nlp(doc.text))
                if w is not None
            ]
            for doc in docs
        ]
        vocabulary = sorted(set(w for sl in content for w in sl))
        cv = CountVectorizer(
            vocabulary=vocabulary, analyzer=lambda x: x  # pyright: ignore
        )
        tfidf = TfidfTransformer()
        self.pipeline = Pipeline([("count", cv), ("tfidf", tfidf)])
        self.pipeline.fit(content)

    def _encode(self, docs: Sequence[Document]) -> List["np.ndarray"]:
        content = [
            [
                w
                for w in (self._word(t) for t in self.nlp(doc.text))
                if w is not None
            ]
            for doc in docs
        ]
        return self.pipeline.transform(content).toarray()

    def _score(
        self, query: str, docs: Sequence[Document]
    ) -> Sequence[Document]:
        query_vec, *_ = self.encode([Document(text=query, did=0)])
        corpus_vec = np.vstack(self.encode(docs))

        scores = np.dot(corpus_vec, query_vec).tolist()
        scored_docs = [d.rank(score=s) for d, s in zip(docs, scores)]

        return scored_docs


class PoolingModes(Enum):
    MEAN = {
        "pooling_mode_cls_token": False,
        "pooling_mode_max_tokens": False,
        "pooling_mode_mean_tokens": True,
        "pooling_mode_mean_sqrt_len_tokens": False,
    }
    MAX = {
        "pooling_mode_cls_token": False,
        "pooling_mode_max_tokens": True,
        "pooling_mode_mean_tokens": False,
        "pooling_mode_mean_sqrt_len_tokens": False,
    }
    CLS = {
        "pooling_mode_cls_token": True,
        "pooling_mode_max_tokens": False,
        "pooling_mode_mean_tokens": False,
        "pooling_mode_mean_sqrt_len_tokens": False,
    }
    SQRT = {
        "pooling_mode_cls_token": False,
        "pooling_mode_max_tokens": False,
        "pooling_mode_mean_tokens": False,
        "pooling_mode_mean_sqrt_len_tokens": True,
    }


@ranker_registry.add("dense")
class SentenceTransformerRanker(BaseRanker["torch.Tensor"]):
    model: "SentenceTransformer"

    def __init__(
        self,
        model_name_or_path: str = "facebook/contriever",
        device: str = "cpu",
        batch_size: int = 512,
        norm_query: bool = False,
        norm_doc: bool = False,
        pooling: Union[None, PoolingModes, str] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        if pooling is None:
            model = SentenceTransformer(model_name_or_path=model_name_or_path)
        else:
            if isinstance(pooling, str):
                pooling = PoolingModes[pooling.upper()]

            enc = models.Transformer(model_name_or_path=model_name_or_path)
            pool = models.Pooling(
                enc.get_word_embedding_dimension(), **pooling.value
            )

            model = SentenceTransformer(modules=[enc, pool])

        self.model = model.to(device)
        self.norm_query = norm_query
        self.norm_doc = norm_doc
        self.batch_size = batch_size

    def _encode(  # type: ignore[override]
        self, batch: Sequence[Document], norm: bool
    ) -> "torch.Tensor":
        encoding = self.model.encode(
            sentences=[doc.text for doc in batch],
            convert_to_tensor=True,
            normalize_embeddings=norm,
            show_progress_bar=False,
            batch_size=self.batch_size,
        )
        assert isinstance(encoding, torch.Tensor)
        return encoding.to("cpu")

    def _score(
        self, query: str, docs: Sequence[Document]
    ) -> Sequence[Document]:
        query_vec, *_ = self.encode(
            [Document(text=query, did=0)], norm=self.norm_doc
        )
        corpus_vec = torch.vstack(self.encode(docs, norm=self.norm_query))

        scores = corpus_vec @ query_vec.reshape(-1, 1)

        scores = scores.squeeze().tolist()
        scored_docs = [d.rank(score=s) for d, s in zip(docs, scores)]

        return scored_docs



import datetime
import json
import multiprocessing
import os
import sys
import time
from pathlib import Path
from time import sleep
from typing import Any, Dict, Literal, NamedTuple, Optional, Union

from shadow_scholar.cli import Argument, cli, safe_import

with safe_import():
    import fire
    import gradio as gr
    import requests
    import torch
    from fairscale.nn.model_parallel.initialize import (  # pyright: ignore
        initialize_model_parallel,
    )
    from llama import (  # pyright: ignore
        LLaMA,
        ModelArgs,
        Tokenizer,
        Transformer,
    )

NUM_GPUS_MAP = {
    "7B": 1,
    "13B": 2,
    "30B": 4,
    "65B": 8,
}


class MpEnv(NamedTuple):
    local_rank: int
    world_size: int

    @classmethod
    def auto(cls):
        return cls(
            local_rank=int(os.environ.get("LOCAL_RANK", -1)),
            world_size=int(os.environ.get("WORLD_SIZE", -1)),
        )


def configure_model_parallel() -> MpEnv:
    env = MpEnv.auto()

    torch.distributed.init_process_group("nccl")  # pyright: ignore
    initialize_model_parallel(env.world_size)
    torch.cuda.set_device(env.local_rank)

    # seed all processes with the same seed
    torch.manual_seed(42)
    return env


def load_model_and_tokenizer(
    llama_dir: Union[str, Path],
    model_name: str,
    env: MpEnv,
    seq_len: int = 1024,
    batch_size: int = 32,
) -> "LLaMA":
    start_load_op = time.time()

    llama_dir = Path(llama_dir)

    if not llama_dir.exists():
        raise ValueError(f"LLaMA directory does not exist: {llama_dir}")

    if model_name not in NUM_GPUS_MAP:
        raise ValueError(f"Invalid model name: {model_name}")

    model_path = llama_dir / model_name

    if not model_path.exists():
        raise ValueError(f"Model path does not exist: {model_path}")

    checkpoints = sorted(model_path.glob("*.pth"))
    if env.world_size != len(checkpoints):
        raise ValueError(
            f"Found {len(checkpoints)} shards, but world is {env.world_size}"
        )

    ckpt_path = checkpoints[env.local_rank]

    print(f"Loading {ckpt_path} to GPU {env.local_rank}...")
    checkpoint = torch.load(str(ckpt_path), map_location="cpu")
    with open(Path(model_path) / "params.json", "r") as f:
        params = json.loads(f.read())

    # create model configuration
    params.update({"max_seq_len": seq_len, "max_batch_size": batch_size})
    model_args = ModelArgs(**params)

    # spin-up tokenizer
    tokenizer = Tokenizer(model_path=str(llama_dir / "tokenizer.model"))

    # create model and set parameters up
    model_args.vocab_size = tokenizer.n_words
    torch.set_default_tensor_type(torch.cuda.HalfTensor)  # pyright: ignore
    model = Transformer(model_args)
    torch.set_default_tensor_type(torch.FloatTensor)
    model.load_state_dict(checkpoint, strict=False)

    # wrap up model in a LLaMA object
    generator = LLaMA(model, tokenizer)

    # complete message
    delta = time.time() - start_load_op
    print(f"Loaded in {delta:.2f} s on GPU {env.local_rank}.")
    return generator


class UI:
    def __init__(
        self,
        model_name: str,
        server_name: str,
        ext_port: int,
        int_port: int,
    ):
        self.model_name = model_name
        self.server_name = server_name
        self.ext_port = ext_port
        self.int_port = int_port

    @property
    def rank(self) -> str:
        if torch.distributed.is_initialized():  # pyright: ignore
            return str(torch.distributed.get_rank())  # pyright: ignore
        else:
            return "null"

    def get(self, what: Literal["input", "output"]) -> dict:
        resp = requests.get(
            f"http://localhost:{self.int_port}/get/{what}?rank={self.rank}"
        )
        return resp.json()

    def set(self, what: Literal["input", "output"], value: dict):
        requests.post(
            f"http://localhost:{self.int_port}/set/{what}?rank={self.rank}",
            json=value,
        )

    def delete(self, what: Literal["input", "output"]):
        requests.get(
            f"http://localhost:{self.int_port}/delete/{what}?rank={self.rank}"
        )

    def runner(
        self, text: str, max_length: int, temperature: float, top_p: float
    ) -> dict:
        self.set(
            "input",
            {
                "text": text,
                "max_length": max_length,
                "temperature": temperature,
                "top_p": top_p,
            },
        )
        output = {}
        while True:
            output = self.get("output")
            if output:
                break
            sleep(1)

        self.delete("output")
        return output["text"]

    def start_ui(self):
        demo = gr.Interface(
            fn=self.runner,
            inputs=[
                gr.Text(lines=10, label="Input"),
                gr.Slider(
                    minimum=1,
                    maximum=2048,
                    step=1,
                    value=256,
                    label="Max Length",
                ),
                gr.Slider(
                    minimum=0.0,
                    maximum=1.0,
                    step=0.05,
                    value=0.8,
                    label="Temperature",
                ),
                gr.Slider(
                    minimum=0.0,
                    maximum=1.0,
                    step=0.05,
                    value=0.95,
                    label="Top P",
                ),
            ],
            outputs=gr.Text(lines=10, label="Output"),
            title=f"LLaMA {self.model_name} Demo",
            # allow_flagging=False,
            # logdir=lg
        )
        demo.queue(concurrency_count=1)
        demo.launch(server_name=self.server_name, server_port=self.ext_port)

    def start_server(self):
        from flask import Flask, jsonify, request

        app = Flask(__name__)

        g: Dict[str, Any] = {"input": None, "output": None}

        def content(what: str) -> dict:
            return {**(g.get(what, None) or {})}

        @app.route("/get/<what>")
        def _get(what: Literal["input", "output"]):
            return jsonify(**content(what))

        @app.route("/set/<what>", methods=["POST"])
        def _set(what: Literal["input", "output"]):
            if request.json is not None:
                g[what] = dict(request.json)
            return jsonify(**content(what))

        @app.route("/delete/<what>")
        def _delete(what: Literal["input", "output"]):
            g[what] = {}
            return jsonify(**content(what))

        app.run()


@cli(
    "app.llama",
    arguments=[
        Argument(
            "-r",
            "--model-root",
            required=True,
            help="Path to directory containing model checkpoints",
        ),
        Argument(
            "-m",
            "--model-name",
            default="7B",
            choices=list(NUM_GPUS_MAP.keys()),
            help="Pretrained model or path to local checkpoint",
        ),
        Argument(
            "-p",
            "--server-port",
            default=7860,
            help="Port to run the server on",
        ),
        Argument(
            "-n",
            "--server-name",
            default="0.0.0.0",
            help="Server address to run the gradio app at",
        ),
        Argument(
            "-l",
            "--logdir",
            default=None,
            help="Directory to log inputs and outputs to",
        ),
    ],
    requirements=[
        "gradio",
        "torch>=1.13",
        "fairscale",
        "fire",
        "sentencepiece",
        "requests",
        "flask",
    ],
)
def run_llama_demo(
    model_root: str,
    model_name: str = "7B",
    server_port: int = 7860,
    server_name: str = "localhost",
    logdir: Optional[Union[str, Path]] = None,
):
    num_gpus = NUM_GPUS_MAP[model_name]

    try:
        import llama  # noqa: F401    # pyright: ignore
    except ImportError:
        msg = (
            "LLaMA is not installed; you have to do so manually. "
            "Please run `pip install git+https://github.com/facebookresearch/"
            "llama.git@76066b1` and try again."
        )
        print(msg, file=sys.stderr)
        sys.exit(1)

    try:
        mp = configure_model_parallel()
    except ValueError:
        # something went wrong with model parallel setup
        mp = MpEnv(-1, -1)

    if mp.world_size < 0:
        message = (
            "This application is meant to be launched with "
            "`torch.distributed.launch`, but it appears that "
            "this is not the case. Please launch the application "
            "with the following command:\n"
            f"torchrun --nproc_per_node {num_gpus} {__file__} "
            f"--model-root {model_root} "
            f"--model-name {model_name} "
            f"--server-port {server_port} "
            f"--server-name {server_name} "
            f"--logdir {logdir}"
        )
        print(message, file=sys.stderr)
        sys.exit(1)

    if logdir:
        current_date = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        logdir = Path(f"{logdir}/{model_name}_{current_date}.jsonl")
        if mp.local_rank == 0:
            logdir.parent.mkdir(parents=True, exist_ok=True)

    ui = UI(
        model_name=model_name,
        server_name=server_name,
        ext_port=server_port,
        int_port=7860,
    )

    ps = []
    if mp.local_rank == 0:
        # start UI and communication server
        ps.append(multiprocessing.Process(target=ui.start_server))
        ps.append(multiprocessing.Process(target=ui.start_ui))

    for p in ps:
        p.start()

    print(f"Starting Llama demo, rank: {mp.local_rank}")

    try:
        # load models
        model_root = model_root.rstrip("/")
        generator = load_model_and_tokenizer(
            llama_dir=model_root,
            model_name=model_name,
            env=mp,
            seq_len=1024,
            batch_size=32,
        )

        torch.distributed.barrier()  # pyright: ignore

        while True:
            input_data = ui.get("input")
            if not input_data:
                sleep(1)
                continue

            if mp.local_rank == 0:
                content = json.dumps(input_data, indent=2)
                print(f"RANK {mp.local_rank}: {content}")

            text = input_data["text"].strip()

            if len(text) > 0:
                results = generator.generate(
                    [text],
                    max_gen_len=int(input_data["max_length"]),
                    temperature=float(input_data["temperature"]),
                    top_p=float(input_data["top_p"]),
                )
            else:
                results = [""]

            output_data = {"text": results[0]}

            if logdir is not None:
                with open(logdir, "a") as f:
                    data = json.dumps(
                        {"input": input_data, "output": output_data}, indent=2
                    )
                    f.write(data + "\n")

            if mp.local_rank == 0:
                data = json.dumps(output_data, indent=2)
                print(f"RANK {mp.local_rank}: {data}")

            if mp.local_rank == 0:
                ui.set("output", output_data)
                ui.delete("input")
                sleep(1)

            torch.distributed.barrier()  # pyright: ignore
    finally:
        for p in ps:
            p.terminate()
            p.join()
        if mp.local_rank == 0:
            gr.close_all()


if __name__ == "__main__":
    fire.Fire(run_llama_demo)



from .app.chat_s2.main import run_v2_demo
from .app.galactica.main import run_galactica_demo
from .app.hello_world.main import run_hello_world
from .app.llama.main import run_llama_demo
from .app.pdod.main import run_pdod, run_pdod_web_ui
from .app.qa.main import run_qa_demo
from .collections.athena import get_s2ag_abstracts
from .hello_world import angry_world, hello_world

__all__ = [
    "angry_world",
    "get_s2ag_abstracts",
    "hello_world",
    "run_galactica_demo",
    "run_hello_world",
    "run_llama_demo",
    "run_pdod_web_ui",
    "run_pdod",
    "run_qa_demo",
    "run_v2_demo",
]



from shadow_scholar.cli import Argument, cli, safe_import

with safe_import():
    import gradio as gr
    import nltk
    import numpy as np
    import openai
    import pandas as pd
    import requests
    import torch
    from transformers import AutoModel, AutoTokenizer


SEARCH_URI = "https://api.semanticscholar.org/graph/v1/paper/search/"
PROMPT_TXT = """\
Answer the question based on the context below.

Context: {context}

Question: {question}

Answer:\
"""


def cosine_similarity(A, B):
    dot_product = A @ B.T
    norm_A = np.linalg.norm(A)
    norm_B = np.linalg.norm(B)
    similarity = dot_product / (norm_A * norm_B)
    return similarity


def search(query, s2_key, limit=20, fields=["title", "abstract"]):
    # space between the  query to be removed and replaced with +
    query = query.replace(" ", "+")
    url = f'{SEARCH_URI}?query={query}&limit={limit}&fields={",".join(fields)}'
    headers = {"Accept": "*/*", "X-API-Key": s2_key}
    response = requests.get(url, headers=headers)
    return response.json()


# function to preprocess the query and remove the stopwords
# before passing it to the search function
def preprocess_query(query):
    query = query.lower()
    # remove stopwords from the query
    stopwords = set(nltk.corpus.stopwords.words("english"))
    query = " ".join([word for word in query.split() if word not in stopwords])
    return query


class SpecterEmbeddings:
    def __init__(self, model_name="allenai/specter"):
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.model = (
            AutoModel.from_pretrained(model_name, torch_dtype=torch.bfloat16)
            .to("cuda" if torch.cuda.is_available() else "cpu")
            .eval()
        )

    def __call__(self, text):
        with torch.inference_mode():
            tokens = self.tokenizer(
                text,
                padding=True,
                truncation=True,
                return_tensors="pt",
                max_length=512,
            ).to(self.model.device)
            embeddings = self.model(**tokens).pooler_output
            return embeddings.detach().to("cpu", dtype=torch.float32).numpy()


def create_context(question, df, max_len=3800, size="davinci"):
    """
    Create a context for a question by finding the most
    similar context from the dataframe
    """

    returns = []
    cur_len = 0

    # Sort by distance and add the text to the context
    # until the context is too long
    for i, row in df.iterrows():
        # Add the length of the text to the current length
        cur_len += row["n_tokens"] + 4

        # If the context is too long, break
        if cur_len > max_len:
            break

        # Else add it to the text that is being returned
        returns.append(row["title_abs"])

    # Return the context
    return "\n\n###\n\n".join(returns)


def answer_question(
    df,
    model="text-davinci-003",
    question="What is the impact of creatine on cognition?",
    max_len=3800,
    size="ada",
    debug=False,
    max_tokens=150,
    stop_sequence=None,
):
    """
    Answer a question based on the most similar context from the
    dataframe texts
    """
    context = create_context(
        question,
        df,
        max_len=max_len,
        size=size,
    )
    # If debug, print the raw model response
    if debug:
        print("Context:\n" + context)
        print("\n\n")

    try:
        # Create a completions using the question and context
        response = openai.Completion.create(
            prompt=PROMPT_TXT.format(context=context, question=question),
            temperature=0,
            max_tokens=max_tokens,
            top_p=1,
            frequency_penalty=0,
            presence_penalty=0,
            stop=stop_sequence,
            model=model,
        )
        return response["choices"][0]["text"].strip()
    except Exception as e:
        print(e)
        return ""


class App:
    def __init__(self, s2_key: str, openai_key: str):
        self.s2_key = s2_key
        self.openai_key = openai_key
        self.embeddings = SpecterEmbeddings()
        openai.api_key = openai_key
        nltk.download("stopwords")

    def __call__(self, query: str) -> str:
        # json to pandas dataframe
        search_results = search(preprocess_query(query), s2_key=self.s2_key)

        if search_results["total"] == 0:
            return "No results found - Try another query"
        else:
            df = pd.DataFrame(search_results["data"]).dropna()

        # merge columns title and abstract into a string separated by
        # tokenizer.sep_token and store it in a list

        df["title_abs"] = [
            d["title"]
            + self.embeddings.tokenizer.sep_token
            + (d.get("abstract") or "")
            for d in df.to_dict("records")
        ]
        df["n_tokens"] = df.title_abs.apply(
            lambda x: len(self.embeddings.tokenizer.encode(x))
        )

        # get embeddings for each document and query
        doc_embeddings = self.embeddings(list(df["title_abs"]))
        query_embeddings = self.embeddings(query)

        df["specter_embeddings"] = list(doc_embeddings)
        # find the cosine similarity between the query and the documents
        df["similarity"] = cosine_similarity(
            query_embeddings, doc_embeddings
        ).flatten()

        # sort the dataframe by similarity
        df.sort_values(by="similarity", ascending=False, inplace=True)

        return answer_question(df, question=query, debug=False)


@cli(
    "app.s2qa",
    arguments=[
        Argument(
            "-sp",
            "--server-port",
            default=7860,
            help="Port to run the server on",
        ),
        Argument(
            "-sn",
            "--server-name",
            default="localhost",
            help="Server address to run the gradio app at",
        ),
        Argument(
            "-ok",
            "--openai-key",
            required=True,
            help="OpenAI API key",
        ),
        Argument(
            "-sk",
            "--s2-key",
            required=True,
            help="Semantic Scholar API key",
        ),
    ],
    requirements=[
        "requests",
        "nltk",
        "transformers",
        "openai",
        "torch",
        "pandas",
        "accelerate",
    ],
)
def run_qa_demo(
    server_port: int,
    server_name: str,
    openai_key: str,
    s2_key: str,
):
    app = App(s2_key=s2_key, openai_key=openai_key)

    demo = gr.Interface(
        title="S2QA Demo",
        fn=app,
        inputs=[
            gr.inputs.Textbox(
                label="Question",
                lines=3,
                placeholder="What is the impact of creatine on cognition?",
            )
        ],
        outputs=[
            gr.Textbox(
                label="Answer",
                lines=3,
            )
        ],
    )

    demo.launch(
        server_port=server_port, server_name=server_name, enable_queue=True
    )



from functools import partial
from typing import (
    Callable,
    Dict,
    Generic,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

T = TypeVar("T")
V = TypeVar("V")


class CallableRegistry(Generic[T]):
    """Singleton registry for datasets."""

    callables: Dict[str, Union[Callable[..., T], Type[T]]]

    def __init__(self):
        self.callables = {}

    @overload
    def _decorator(  # type: ignore
        self, func: Callable[..., T], func_name: Optional[str] = None
    ) -> Callable[..., T]:
        ...

    @overload
    def _decorator(
        self, func: Type[T], func_name: Optional[str] = None
    ) -> Type[T]:
        ...

    def _decorator(
        self,
        func: Union[Callable[..., T], Type[T]],
        func_name: Optional[str] = None,
    ) -> Union[Callable[..., T], Type[T]]:
        name = func_name or func.__name__
        if name in self.callables:
            raise KeyError(f"Dataset {name} already in the registry")
        self.callables[name] = func
        return func

    def add(self, name: str) -> Callable[[V], V]:
        return partial(self._decorator, func_name=name)  # type: ignore

    def get(self, name: str) -> Callable[..., T]:
        return self.callables[name]

    def keys(self) -> Tuple[str, ...]:
        return tuple(self.callables.keys())



import gzip
import json
from dataclasses import dataclass, field
from functools import partial
from multiprocessing import Manager, Pool, cpu_count
from queue import Queue
from threading import Thread
from typing import Iterable, List, Literal, Optional, Union

from shadow_scholar.cli import Argument, cli, safe_import

with safe_import():
    from smashed.utils.io_utils import (
        MultiPath,
        open_file_for_read,
        open_file_for_write,
        recursively_list_files,
    )
    from tqdm import tqdm


@dataclass
class Section:
    header: Optional[str] = None
    paragraphs: List[str] = field(default_factory=list)

    def __str__(self) -> str:
        return (f"{self.header}\n\n" if self.header else "") + "\n\n".join(
            self.paragraphs
        )

    def split(self) -> List["Section"]:
        return [
            Section(header=self.header, paragraphs=[p])
            for p in self.paragraphs
        ]

    @property
    def valid(self) -> bool:
        return len(self.paragraphs) > 0


def _process_document(
    doc: dict,
    unit: Union[Literal["paragraph"], Literal["document"]],
) -> Iterable[dict]:
    if (metadata := doc.get("metadata", None)) is None:
        return

    if (title := metadata.get("title", None)) is None:
        return

    if (abstract := metadata.get("abstract", None)) is None:
        return

    if (authors := metadata.get("authors", None)) is None:
        return
    authors = json.loads(authors)

    if (venue := metadata.get("venue", None)) is None:
        return

    if (publication_date := metadata.get("publication_date", None)) is None:
        return
    year = publication_date["year"]

    if (external_ids := metadata.get("external_ids", None)) is None:
        return

    ids = {
        "acl": external_ids.get("acl", None),
        "arxiv": external_ids.get("arxiv", None),
        "doi": external_ids.get("doi", None),
        "corpus": doc["id"],
    }
    tldr = doc["tldr"]

    sections: List[Section] = []

    if doc.get("content", None) is not None:
        # content down here
        raw = (grobid := doc["content"]["grobid"])["contents"]

        heads = json.loads(grobid["annotations"]["section_header"] or "[]")
        paras = json.loads(grobid["annotations"]["paragraph"] or "[]")
        secs_it = (
            (name, loc)
            for name, locs in (("header", heads), ("paragraph", paras))
            for loc in locs
        )

        current_section = Section()

        for name, loc in sorted(secs_it, key=lambda x: x[1]["start"]):
            content = raw[int(loc["start"]) : int(loc["end"])]

            if name == "header":
                if current_section.valid:
                    sections.append(current_section)
                current_section = Section(header=content)
                continue
            current_section.paragraphs.append(content)

        if current_section.paragraphs:
            # add the last section
            sections.append(current_section)

    base = {
        "title": title,
        "authors": authors,
        "venue": venue,
        "year": year,
        "ids": ids,
        "tldr": tldr,
    }

    if unit == "paragraph":
        yield {**base, "paragraph": str(Section("Abstract", [abstract]))}
        for section in sections:
            for paragraph in section.split():
                yield {**base, "paragraph": str(paragraph)}
    elif unit == "document":
        yield {
            **base,
            "abstract": abstract,
            "full_text": [str(sec) for sec in sections],
        }


def _process_collection(
    src: str,
    dst: str,
    unit: Union[Literal["paragraph"], Literal["document"]],
    compressed: bool = True,
    pbar_queue: Optional[Queue] = None,
):
    with open_file_for_read(
        src, "rb" if compressed else "r"
    ) as inf, open_file_for_write(dst, "wb" if compressed else "rb") as outf:
        if compressed:
            inf = gzip.GzipFile(fileobj=inf, mode="rb")
        if compressed:
            outf = gzip.GzipFile(fileobj=outf, mode="wb")

        for ln in inf:
            doc = json.loads(ln)
            for item in _process_document(doc, unit):
                item_str = json.dumps(item) + "\n"
                if compressed:
                    item_str = item_str.encode("utf-8")  # type: ignore
                outf.write(item_str)  # pyright: ignore

            if pbar_queue is not None:
                pbar_queue.put(1)

        outf.close()


def threaded_progressbar(q: Queue, timeout: float):
    with tqdm(desc="Processing...", unit=" doc", unit_scale=True) as pbar:
        while True:
            item = q.get()
            if item is None:
                break
            pbar.update(1)


@cli(
    name="app.qa.collection",
    arguments=[
        Argument(
            "-s",
            "--src",
            type=str,
            required=True,
            help="Path to the collection to process.",
        ),
        Argument(
            "-d",
            "--dst",
            type=str,
            required=True,
            help="Path to the output directory.",
        ),
        Argument(
            "-u",
            "--unit",
            type=str,
            choices=["paragraph", "document"],
            default="paragraph",
            help="Unit to process.",
        ),
        Argument(  # pyright: ignore
            "-D",
            "--Debug",
            dest="debug",
            action="store_true",
            help="Enable debug mode.",
        ),
    ],
    requirements=[
        "smashed[remote]",
    ],
)
def process_collection(
    src: str,
    dst: str,
    unit: Union[Literal["paragraph"], Literal["document"]],
    debug: bool = False,
):
    files = recursively_list_files(src, True)  # pyright: ignore

    if debug:
        for src_fp in files:
            dst_fp = MultiPath.parse(dst) / (src_fp - src)
            _process_collection(src=str(src_fp), dst=str(dst_fp), unit=unit)

    else:
        with Pool(processes=cpu_count()) as pool:
            pbar_queue: Queue = (manager := Manager()).Queue()
            pbar_thread = Thread(
                target=threaded_progressbar,
                args=(pbar_queue, 0.1),
                daemon=True,
            )
            pbar_thread.start()

            results = []
            for src_fp in files:
                dst_fp = MultiPath.parse(dst) / (src_fp - src)
                result = pool.apply_async(
                    partial(
                        _process_collection,
                        src=str(src_fp),
                        dst=str(dst_fp),
                        unit=unit,
                        pbar_queue=pbar_queue,
                    )
                )
                results.append(result)

            pool.close()
            pool.join()

            try:
                while len(results) > 0:
                    results.pop().get()
            except Exception as e:
                raise e
            finally:
                pbar_queue.put(None)
                pbar_thread.join()
                manager.shutdown()



import logging
import re
from ast import literal_eval
from contextlib import contextmanager
from typing import Dict, Iterator, List, Literal, Optional, Tuple, Union

from shadow_scholar.cli import safe_import

with safe_import():
    import torch
    from accelerate import infer_auto_device_map, init_empty_weights
    from transformers import (
        AutoModelForCausalLM,
        AutoTokenizer,
        OPTConfig,
        OPTForCausalLM,
    )


LOGGING = logging.getLogger(__name__)


def get_submodule(model: "torch.nn.Module", target: str) -> "torch.nn.Module":
    ...
    for part in target.split("."):
        if re.match(r"\d+", part):
            model = model[int(part)]  # pyright: ignore
        else:
            model = getattr(model, part)
    return model


def set_submodule(
    model: "torch.nn.Module", target: str, value: "torch.nn.Module"
):
    ...
    parts = target.split(".")
    for part in parts[:-1]:
        if re.match(r"\d+", part):
            model = model[int(part)]  # pyright: ignore
        else:
            model = getattr(model, part)
    setattr(model, parts[-1], value)


def move_tensors(
    module: "torch.nn.Module",
    input: tuple,
    output: Union[Tuple["torch.Tensor", ...], "torch.Tensor"],
    device: str,
) -> Union[Tuple["torch.Tensor", ...], "torch.Tensor"]:
    if isinstance(output, torch.Tensor):
        return output.to(device)
    return tuple(
        move_tensors(module, input, o, device)  # pyright: ignore
        for o in output
    )


def _gpu_mem() -> Dict[int, int]:
    """Return the amount of memory available on GPUs in MB, rounded to the
    nearest integer."""
    return {
        gid: round(
            torch.cuda.get_device_properties(gid).total_memory / 1024 / 1024
        )
        for gid in range(torch.cuda.device_count())
    }


class _gl_model:
    @property
    def variants(cls) -> Dict[str, Tuple[str, "torch.dtype"]]:
        return {
            "mini": ("facebook/galactica-125m", torch.float32),
            "base": ("facebook/galactica-1.3b", torch.float32),
            "standard": ("facebook/galactica-6.7b", torch.float32),
            "large": ("facebook/galactica-30b", torch.float32),
            "huge": ("facebook/galactica-120b", torch.float16),
        }

    def __str__(self):
        return (
            f"model `{self.backbone}` on `{self.device}` "
            f"with type `{self.dtype}`"
        )

    def __init__(
        self,
        model_name: str,
        model_path: Optional[str] = None,
        precision: Literal["full", "mixed", "int8"] = "full",
        use_accelerate: bool = False,
    ):
        self.device = torch.device(
            "cuda" if torch.cuda.is_available() else "cpu"
        )

        assert model_name in self.variants, (
            f"Unknown model name: {model_name}. "
            f"Available models: {', '.join(self.variants.keys())}"
        )
        self.backbone, self.dtype = self.variants[model_name]

        if precision == "full":
            if self.dtype in {torch.float16, torch.int8, torch.bfloat16}:
                LOGGING.warning(
                    f"Model is only available in {self.dtype}. "
                    f"Ignoring `precision={precision}` argument."
                )
        elif precision == "mixed":
            if self.dtype in {torch.int8, torch.float16}:
                LOGGING.warning(
                    f"Model is only available in {self.dtype}. "
                    f"Ignoring `precision={precision}` argument."
                )
            self.dtype = torch.bfloat16
        elif precision == "int8":
            raise NotImplementedError("Int8 is not supported yet.")
        else:
            raise ValueError(f"Unknown precision: {precision}")

        print(f"Loading {self}...", end="", flush=True)

        self.tokenizer = AutoTokenizer.from_pretrained(
            pretrained_model_name_or_path=self.backbone,
        )

        # from the galactica repo: https://github.com/paperswithcode/galai/blob/1a1e4486fd38c85c6b66b3c0b31ce24b0afd7613/galai/model.py#L154  # noqa: E501

        # setup padding
        self.tokenizer.pad_token_id = 1
        self.tokenizer.pad_token = "<pad>"
        self.tokenizer.padding_side = "left"

        # setup truncation
        self.tokenizer.truncation_side = "left"

        # setup special tokens
        self.tokenizer.bos_token_id = 0
        self.tokenizer.bos_token = "<s>"

        self.tokenizer.eos_token_id = 2
        self.tokenizer.eos_token = "</s>"

        self.tokenizer.unk_token = "<unk>"
        self.tokenizer.unk_token_id = 3

        if use_accelerate:
            config = OPTConfig.from_pretrained(self.backbone)
            with init_empty_weights():
                empty_model = AutoModelForCausalLM.from_config(config)

            device_map = infer_auto_device_map(
                empty_model,
                max_memory={
                    gid: f"{mem * .7:.0f}MB" for gid, mem in _gpu_mem().items()
                },
                no_split_module_classes=[
                    "OPTDecoderLayer",
                    "Linear" "Embedding",
                    "OPTLearnedPositionalEmbedding" "LayerNorm",
                ],
                dtype=self.dtype,
            )

            # HACK: we actually want to have "lm_head" on the first GPU
            #       because its weights are tied to the embedding layer
            #       and we want to keep them on the same device.
            if "lm_head" in device_map:
                device_map["lm_head"] = 0
            assert (
                device_map.get(
                    "model.decoder.embed_tokens",
                    0,
                )
                == device_map.get(
                    "model.decoder.embed_positions",
                    0,
                )
                == device_map.get("model.decoder.final_layer_norm", 0)
                == device_map["lm_head"]
            ), "Embedding layers and lm_head should be on the same device."

            self.model = OPTForCausalLM.from_pretrained(
                pretrained_model_name_or_path=model_path or self.backbone,
                device_map=device_map,
                offload_folder="offload",
                offload_state_dict=True,
                torch_dtype=self.dtype,
            ).eval()  # pyright: ignore

        else:
            self.model = (
                OPTForCausalLM.from_pretrained(
                    pretrained_model_name_or_path=model_path or self.backbone,
                    torch_dtype=self.dtype,
                )
                .to(self.device)  # pyright: ignore
                .eval()
            )

        print("done.", flush=True)

    @contextmanager
    def autocast(self) -> Iterator[bool]:
        try:
            if self.device == "cuda" and self.dtype != torch.float32:
                with torch.cuda.amp.autocast(enabled=True):  # pyright: ignore
                    yield True
            else:
                yield False
        finally:
            pass

    def __call__(
        self,
        text: str,
        tokenize_config: Union[dict, List[Tuple[str, str]], None] = None,
        generate_config: Union[dict, List[Tuple[str, str]], None] = None,
    ) -> str:
        if not isinstance(tokenize_config, dict):
            tokenize_config = {
                k: literal_eval(v)
                for k, v in (tokenize_config or [])
                if k.strip() != "" and v.strip() != ""
            }

        if not isinstance(generate_config, dict):
            generate_config = {
                k: literal_eval(v)
                for k, v in (generate_config or [])
                if k.strip() != "" and v.strip() != ""
            }

        batch = self.tokenizer(text, return_tensors="pt", **tokenize_config)
        casted_batch = batch.to(self.device)

        with self.autocast(), torch.no_grad():
            outputs = self.model.generate(
                input_ids=casted_batch.input_ids,
                attention_mask=casted_batch.attention_mask,
                **generate_config,
            )

        # following instructions from the galactica repo: https://github.com/paperswithcode/galai/blob/1a1e4486fd38c85c6b66b3c0b31ce24b0afd7613/galai/model.py#L308     # noqa: E501
        decoded = self.tokenizer.decode(
            outputs[0],
            skip_special_tokens=False,
            clean_up_tokenization_spaces=False,
        )
        decoded = decoded.replace(self.tokenizer.eos_token, "").replace(
            self.tokenizer.pad_token, ""
        )

        return decoded
